import pytest
import numpy as np
import pandas as pd
import pytafast
talib = pytest.importorskip("talib")

# Generate a large dataset for meaningful benchmarking
np.random.seed(42)
LARGE_ARRAY = np.random.random(100_000) * 100
LARGE_SERIES = pd.Series(LARGE_ARRAY)

TIME_PERIOD = 30

def test_benchmark_pytafast_sma_numpy(benchmark):
    benchmark(pytafast.SMA, LARGE_ARRAY, timeperiod=TIME_PERIOD)

def test_benchmark_official_talib_sma_numpy(benchmark):
    benchmark(talib.SMA, LARGE_ARRAY, timeperiod=TIME_PERIOD)

def test_benchmark_pytafast_sma_pandas(benchmark):
    benchmark(pytafast.SMA, LARGE_SERIES, timeperiod=TIME_PERIOD)

def test_benchmark_official_talib_sma_pandas(benchmark):
    benchmark(talib.SMA, LARGE_SERIES, timeperiod=TIME_PERIOD)



import pytest
import asyncio
import tracemalloc
import numpy as np
import pytafast
import gc

def test_high_concurrency_no_leak():
    asyncio.run(_test_high_concurrency_no_leak())

async def _test_high_concurrency_no_leak():
    np.random.seed(42)
    in_real = np.random.random(5000) * 100 + 10
    
    # Pre-warm function to resolve lazy loads (e.g., pandas imports inside wrappers)
    await pytafast.aio.SMA(in_real, timeperiod=14)
    await pytafast.aio.MACD(in_real, fastperiod=12, slowperiod=26, signalperiod=9)
    await pytafast.aio.BBANDS(in_real, timeperiod=5)
    
    # Setup tracemalloc to watch memory AFTER lazy loading
    tracemalloc.start()
    
    # We will simulate high concurrency by running 1000 tasks concurrently
    num_tasks = 1000
    
    # Define an async worker
    async def worker():
        # Mix of single output and tuple output indications
        await pytafast.aio.SMA(in_real, timeperiod=14)
        await pytafast.aio.MACD(in_real, fastperiod=12, slowperiod=26, signalperiod=9)
        await pytafast.aio.BBANDS(in_real, timeperiod=5)
        
    # Take a memory snapshot before
    gc.collect()
    snap_before = tracemalloc.take_snapshot()
    
    # Execute batch concurrency
    tasks = [worker() for _ in range(num_tasks)]
    await asyncio.gather(*tasks)
    
    # Take memory snapshot after
    gc.collect()
    snap_after = tracemalloc.take_snapshot()
    
    # Get top stats (differences)
    stats = snap_after.compare_to(snap_before, 'lineno')
    tracemalloc.stop()
    
    # Ignore negligible small allocations (like Python internal runtime diffs).
    # We specifically want to check if the massive NumPy arrays or capsules are leaking.
    # We sum the difference in bytes.
    total_diff_bytes = sum(stat.size_diff for stat in stats)
    
    # 5MB threshold for safety, usually a real leak of 1000 tasks doing TA-lib arrays would be hundreds of MBs
    assert total_diff_bytes < 5_000_000, f"Memory leak detected! Total difference: {total_diff_bytes / 1024 / 1024:.2f} MB"



#!/usr/bin/env python3

# Produces the assets release candidates in 'dist'.
#
# The outputs depend of the host system.
#
#    For linux/ubuntu: ta-lib-<version>-src.tar.gz
#         with contents for doing "./configure; make; sudo make install"
#
#    For windows: ta-lib-<version>-windows-<platform>.zip
#         with contents:
#            bin/ta-lib.dll        (dynamic library)
#            lib/ta-lib.lib        (import library)
#            lib/ta-lib-static.lib (static library)
#            include/*.h           (API headers)
#            VERSION.txt           Version number "major.minor.patch"
#
# How to run it?
#   Do './scripts/package.py' while current directory is the root of the ta-lib repository.
#
#   Windows Specific:
#    - You must have Visual Studio installed (free community version works).
#    - Host machine *must* be x64.
#
#    (FYI, all this can optionally be done in a Windows VM)
#
# How to change the version?
#   Edit MAJOR, MINOR, PATCH in src/ta_common/ta_version.c
#   There is no need to modify other files (they will be updated by this script).
#
#   See README-DEVS.md for all the release steps.

import argparse
import copy
from multiprocessing.context import _force_start_method
import os
import shlex
import subprocess
import sys
import glob
import platform
import shutil

from utilities.package_digest import PackageDigest
from utilities.windows import call_vcvarsall
from utilities.versions import sync_sources_digest, sync_versions
from utilities.common import are_generated_files_git_changed, compare_dir, copy_file_list, create_temp_dir, get_git_bot_user_name, get_git_user_name, get_src_generated_files, is_arm64_toolchain_installed, is_cmake_installed, is_debian_based, is_dotnet_installed, is_i386_toolchain_installed, is_msbuild_installed, is_nightly_github_action, is_redhat_based, is_rpmbuild_installed, is_ubuntu, is_dotnet_installed, is_wix_installed, is_x86_64_toolchain_installed, run_command, run_command_term, verify_git_repo, run_command_sudo
from utilities.files import compare_msi_files, compare_tar_gz_files, compare_zip_files, create_rtf_from_txt, create_zip_file, compare_deb_files, force_delete, force_delete_glob, path_join

def delete_other_versions(target_dir: str, file_pattern: str, new_version: str ):
    # Used for cleaning-up a directory from other versions than the one
    # being newly built.
    #
    # Example of file_pattern: 'ta-lib-*-src.tar.gz'
    glob_all_packages = path_join(target_dir, file_pattern)
    for file in glob.glob(glob_all_packages):
        if new_version not in file:
            force_delete(file)

def do_cmake_reconfigure(root_dir: str, options: str, sudo_pwd: str = "") ->str:
    # Clean-up any potential previous build, and configure a new build
    # using the CMakeLists.txt and specified options.
    #
    # Returns the "build" location where further cmake/cpack should be done from.
    #
    # Exit on any error.
    build_dir = path_join(root_dir, 'build')
    if os.path.exists(build_dir):
        force_delete(build_dir, sudo_pwd)

    if not is_cmake_installed():
        print("Error: CMake not found. You need to install it and be accessible with PATH.")
        sys.exit(1)

    cmake_lists = path_join(root_dir, 'CMakeLists.txt')
    if not os.path.isfile(cmake_lists):
        print(f"Error: {cmake_lists} not found. Your working directory must be within a cloned TA-Lib repos")
        sys.exit(1)

    # Delete ta_config.h to force regeneration.
    ta_config_h = path_join(root_dir, 'include', 'ta_config.h')
    force_delete(ta_config_h, sudo_pwd)

    # Run CMake configuration.
    original_dir = os.getcwd()
    try:
        os.makedirs(build_dir)
        os.chdir(build_dir)
        cmake_command = ['cmake'] + shlex.split(options) + ['..']
        formatted_command = ' '.join([f'[{elem}]' for elem in cmake_command])
        print("CMake configure command:", formatted_command)  # to help debugging, display each arg in brackets
        run_command_term(cmake_command) # Run the command in foreground (output displayed).
    finally:
        # Restore the original working directory
        os.chdir(original_dir)

    return build_dir

def do_cmake_build(build_dir: str):
    # Exit on any error.
    original_dir = os.getcwd()
    try:
        os.chdir(build_dir)
        cmake_command = ['cmake', '--build', '.']
        formatted_command = ' '.join([f'[{elem}]' for elem in cmake_command])
        print("CMake build command:", formatted_command)  # to help debugging, display each arg in brackets
        run_command(cmake_command) # Output piped and displayed only on error.
    finally:
        # Restore the original working directory
        os.chdir(original_dir)

def do_cpack_build(build_dir: str):
    # Exit on any error.
    original_dir = os.getcwd()
    try:
        os.chdir(build_dir)
        cpack_command = ['cpack', '.']
        formatted_command = ' '.join([f'[{elem}]' for elem in cpack_command])
        print("CPack command:", formatted_command)  # to help debugging, display each arg in brackets
        run_command_term(cpack_command)
    finally:
        # Restore the original working directory
        os.chdir(original_dir)

def find_asset_with_ext(target_dir, version: str, extension: str) -> str:
    # Useful for identifying the name of the file generated by CMake.
    # Exit on error.
    glob_deb = path_join(target_dir, f"*.{extension}")
    files = glob.glob(glob_deb)
    if len(files) != 1:
        print(f"Error: Expected one .{extension} file, found {len(files)}")
        sys.exit(1)

    # Check that the single file has the version as substring
    filepath = files[0]
    if version not in filepath:
        print(f"Error: Expected version {version} not found in {filepath}")
        sys.exit(1)

    return os.path.basename(filepath)

def package_windows_zip(root_dir: str, asset_file_name: str, version: str, sources_digest: str, builder_id: str) -> dict:
    result: dict = {"build_valid": False}
    result["asset_file_name"] = asset_file_name

    # Validate the asset_file_name
    if not asset_file_name.endswith('.zip'):
        print(f"Error: Invalid asset_file_name {asset_file_name}. Expected a .zip file.")
        return
    if version not in asset_file_name:
        print(f"Error: Invalid asset_file_name {asset_file_name}. Expected version {version}.")
        return

    file_name_prefix = asset_file_name[:-4]

    # Check dependencies.
    if not is_msbuild_installed():
        print("Error: MSBuild not found. It is required to build the package.")
        sys.exit(1)

    # Clean-up
    dist_dir = path_join(root_dir, 'dist')
    delete_other_versions(dist_dir,"ta-lib-*.zip",version)

    temp_dir = path_join(root_dir, 'temp')
    delete_other_versions(temp_dir,"ta-lib-*.zip",version)

    force_delete_glob(temp_dir, "ta-lib-*")

    # Delete previous package digest files.
    digests_dir = os.path.join(dist_dir, 'digests')
    delete_other_versions(digests_dir, "*.digest", version)

    if is_build_skipping_allowed(root_dir, asset_file_name, version, sources_digest, builder_id):
        result["build_valid"] = True
        result["existed"] = True
        result["copied"] = False
        return result

    # Build the libraries
    build_dir = do_cmake_reconfigure(root_dir, '-G Ninja -DBUILD_DEV_TOOLS=OFF -DCMAKE_BUILD_TYPE=Release')
    do_cmake_build(build_dir)

    # Create a temporary zip package to test before copying to dist.
    package_temp_dir = path_join(temp_dir, file_name_prefix)
    temp_lib_dir = path_join(package_temp_dir, 'lib')
    temp_bin_dir = path_join(package_temp_dir, 'bin')
    temp_include_dir = path_join(package_temp_dir, 'include')

    os.makedirs(temp_dir, exist_ok=True)
    os.makedirs(package_temp_dir, exist_ok=True)
    os.makedirs(temp_lib_dir, exist_ok=True)
    os.makedirs(temp_bin_dir, exist_ok=True)
    os.makedirs(temp_include_dir, exist_ok=True)

    # Copy the built files to the temporary package locations.
    original_dir = os.getcwd()
    include_rootdir = path_join(root_dir, 'include')
    try:
        os.chdir(build_dir)
        shutil.copy('ta-lib.dll', temp_bin_dir)
        shutil.copy('ta-lib.lib', temp_lib_dir)
        shutil.copy('ta-lib-static.lib', temp_lib_dir)
        for header_filepath in glob.glob(path_join(include_rootdir, '*.h')):
            if "ta_config.h" in header_filepath:
                # At this point, this build artifact header is no longuer needed.
                force_delete(header_filepath)
                continue
            shutil.copy(header_filepath, temp_include_dir)

        # Create the VERSION.txt file
        with open(path_join(package_temp_dir, 'VERSION.txt'), 'w') as f:
            f.write(version)
    except subprocess.CalledProcessError as e:
        print(f"Error copying files: {e}")
        return result
    finally:
        # Restore the original working directory
        os.chdir(original_dir)

    # Compress the package (.zip)
    package_temp_file = path_join(temp_dir, asset_file_name)
    try:
        create_zip_file(package_temp_dir, package_temp_file)
    except subprocess.CalledProcessError as e:
        print(f"Error creating zip file: {e}")
        return

    # TODO Add some real "end-user installation" testing. Now just pretend is is OK...
    result["dist_test_pass"] = True

    # Temporary zip is verified OK, so copy it into dist, but only if its content is different.
    os.makedirs(dist_dir, exist_ok=True)
    dist_file = path_join(dist_dir, asset_file_name)
    package_existed = os.path.exists(dist_file)
    package_copied = False
    if not package_existed or not compare_zip_files(package_temp_file, dist_file):
        os.makedirs(dist_dir, exist_ok=True)
        shutil.copy(package_temp_file, dist_file)
        package_copied = True

    result["build_valid"] = True
    result["existed"] = package_existed
    result["copied"] = package_copied
    return result

def package_windows_msi(root_dir: str, asset_file_name: str, version: str, sources_digest: str, builder_id: str, force_build: bool) -> dict:
    result: dict = {"build_valid": False}
    result["asset_file_name"] = asset_file_name

    # Validate the asset_file_name
    if not asset_file_name.endswith('.msi'):
        print(f"Error: Invalid asset_file_name {asset_file_name}. Expected a .msi file.")
        return
    if version not in asset_file_name:
        print(f"Error: Invalid asset_file_name {asset_file_name}. Expected version {version}.")
        return

    # Check dependencies.
    if not is_msbuild_installed():
        print("Error: MSBuild not found. It is required to build the package.")
        sys.exit(1)

    if not is_dotnet_installed():
       print("Error: .NET Framework not found. It is required to build the MSI.")
       return result

    if not is_wix_installed():
        print("Error: WiX Toolset not found. It is required to build the MSI.")
        return result

    # Clean-up
    dist_dir = path_join(root_dir, 'dist')
    delete_other_versions(dist_dir,"ta-lib-*.msi",version)

    temp_dir = path_join(root_dir, 'temp')
    force_delete(path_join(temp_dir, "PFiles"))

    force_delete_glob(temp_dir, "ta-lib-*")

    # Delete previous package digest files.
    digests_dir = os.path.join(dist_dir, 'digests')
    delete_other_versions(digests_dir, "*.digest", version)

    if not force_build and is_build_skipping_allowed(root_dir, asset_file_name, version, sources_digest, builder_id):
        result["build_valid"] = True
        result["existed"] = True
        result["copied"] = False
        return result

    # MSI supports only .rtf for license, so generate it into root_dir.
    license_txt = path_join(root_dir,"LICENSE")
    license_rtf = path_join(root_dir,"LICENSE.rtf")
    create_rtf_from_txt(license_txt,license_rtf)

    build_dir = do_cmake_reconfigure(root_dir, '-G Ninja -DCPACK_GENERATOR=WIX -DBUILD_DEV_TOOLS=OFF -DCMAKE_BUILD_TYPE=Release')
    do_cmake_build(build_dir) # Build the libraries
    do_cpack_build(build_dir) # Create the .msi

    built_asset_file_name = find_asset_with_ext(build_dir, version, "msi")
    if built_asset_file_name != asset_file_name:
        print(f"Error: Expected file name {asset_file_name}, but got {built_asset_file_name}")
        return result

    # TODO Add some real "end-user installation" testing. Now just pretend is is OK...
    result["dist_test_pass"] = True

    # Temporary msi is verified OK, so copy it into dist, but only if its content is different.
    temp_dist_file = path_join(build_dir, asset_file_name)
    dist_dir = path_join(root_dir, 'dist')
    dist_file = path_join(dist_dir, asset_file_name)
    package_existed = os.path.exists(dist_file)
    package_copied = False
    if force_build or not package_existed or not compare_msi_files(temp_dist_file, dist_file):
        os.makedirs(dist_dir, exist_ok=True)
        if os.path.exists(dist_file):
            os.remove(dist_file)
        os.rename(temp_dist_file, dist_file)
        package_copied = True

    result["build_valid"] = True
    result["existed"] = package_existed
    result["copied"] = package_copied
    return result

def package_deb(root_dir: str, asset_file_name: str, version: str, sources_digest: str, builder_id: str, sudo_pwd: str, toolchain: str, force_build: bool) -> dict:
    # Create .deb packaging to be installed with apt or dpkg (Debian-based systems).
    #
    # TA-Lib will install under '/usr/lib' and '/usr/include/ta-lib'.
    #
    # The asset is created/updated into the 'dist' directory only when it
    # pass some tests.
    #
    result: dict = {"build_valid": False}
    result["asset_file_name"] = asset_file_name

    # Check dependencies.
    if not is_debian_based():
        print("Error: Debian-based system required for .deb packaging.")
        return result

    if not is_cmake_installed():
        print("Error: CMake not found. It is required to be install for .deb creation")
        return result

    cmakelists = os.path.join(root_dir, 'CMakeLists.txt')
    if not os.path.isfile(cmakelists):
        print(f"Error: {cmakelists} not found. Make sure you are working within a TA-Lib repos")
        return result

    # Clean-up
    dist_dir = path_join(root_dir, 'dist')
    delete_other_versions(dist_dir,"*.deb",version)

    # Delete previous package digest files.
    digests_dir = os.path.join(dist_dir, 'digests')
    delete_other_versions(digests_dir, "*.digest", version)

    if not force_build and is_build_skipping_allowed(root_dir, asset_file_name, version, sources_digest, builder_id):
        result["build_valid"] = True
        result["existed"] = True
        result["copied"] = False
        return result

    # Build the libraries
    configure_options = '-DCPACK_GENERATOR=DEB -DBUILD_DEV_TOOLS=OFF'

    if toolchain:
        cmake_dir = path_join(root_dir, 'cmake')
        toolchain_file = path_join(cmake_dir, toolchain)
        configure_options += f' -DCMAKE_TOOLCHAIN_FILE={toolchain_file}'

    build_dir = do_cmake_reconfigure(root_dir, configure_options, sudo_pwd)
    do_cmake_build(build_dir)
    do_cpack_build(build_dir)

    # Get the asset file name (from the only .deb file expected in the build directory)
    glob_deb = os.path.join(build_dir, '*.deb')
    deb_files = glob.glob(glob_deb)
    if len(deb_files) != 1:
        print(f"Error: Expected one .deb file, found {len(deb_files)}")
        return result

    # Sanity check that it matches the expected asset_file_name.
    deb_file = deb_files[0]
    built_file_name = os.path.basename(deb_file)
    if built_file_name != asset_file_name:
        print(f"Error: Expected file name [{asset_file_name}] but got [{built_file_name}]")
        return result

    # Sanity check that the expected asset exists.
    test_file_path = os.path.join(build_dir, asset_file_name)
    if not os.path.exists(test_file_path):
        print(f"Error: {test_file_path} not found.")
        return result

    # TODO Add some real "end-user installation" testing. Now just pretend is is OK...
    result["dist_test_pass"] = True

    # Copy the .deb file into dist, but only if it is binary different
    # The creation date will be ignored.
    dist_file = os.path.join(dist_dir, asset_file_name)
    package_existed = os.path.exists(dist_file)
    package_copied = False
    if force_build or not package_existed or not compare_deb_files(deb_file, dist_file):
        os.makedirs(dist_dir, exist_ok=True)
        os.rename(deb_file, dist_file)
        package_copied = True

    result["build_valid"] = True
    result["existed"] = package_existed
    result["copied"] = package_copied
    return result

def package_rpm(root_dir: str, version: str, sudo_pwd: str) -> dict:
    result: dict = {"build_valid": False}
    # Not implemented yet
    return result


def is_build_skipping_allowed(root_dir: str, asset_file_name: str, version: str, sources_digest: str, builder_id: str) -> bool:
    # Check if possible to skip the rebuild and tests.
    #
    # Only allowed when the package digest confirms the binaries have
    # not changed AND all enabled tests already pass for it.
    dist_dir = os.path.join(root_dir, 'dist')

    # Check if the digest file exists.
    digests_dir = os.path.join(dist_dir, 'digests')
    digest_file = os.path.join(digests_dir, f"{asset_file_name}.digest")
    if not os.path.exists(digest_file):
        print(f"Info: Digest file not found {digest_file}.")
        return False

    try:
        pdigest = PackageDigest.read_or_create(root_dir, asset_file_name, sources_digest, builder_id )
        if pdigest.built_success and pdigest.sources_digest == sources_digest:
            # Double-check that the dist/ binaries still matches the MD5.
            local_asset_md5 = pdigest.calculate_md5()
            if local_asset_md5 == "Disabled":
                print("Error: MD5 unexpectadly disabled for digest file.")
                sys.exit(1)
            if local_asset_md5 == pdigest.package_md5 and pdigest.are_all_tests_passed():

                if os.getenv('GITHUB_ACTIONS') == 'true':
                    if is_nightly_github_action():
                      print(f"Info: {asset_file_name} will be rebuild for scheduled github action")
                      return False

                    if pdigest.builder_id != get_git_bot_user_name():
                      print(f"Info: {asset_file_name} from {pdigest.builder_id} will be rebuild by {get_git_bot_user_name()}")
                      return False

                print(f"Info: Allow skipping of already built and tested {asset_file_name}")
                return True
    except Exception as e:
        print(f"Warning: Could not read digest file for {asset_file_name}: {e}")

    return False


def package_src_tar_gz(root_dir: str, asset_file_name: str, version: str, sources_digest: str, builder_id: str, sudo_pwd: str) -> dict:
    # The src.tar.gz is for users wanting to build from source with autotools (./configure).
    #
    # The asset is created/updated into the 'dist' directory only as needed and passing
    # some basic tests.
    #
    result: dict = {"processed": True, "build_valid": False}
    result["asset_file_name"] = asset_file_name

    dist_dir = os.path.join(root_dir, 'dist')

    # Delete previous dist packaging
    delete_other_versions(dist_dir, "*-src.tar.gz", version)

    temp_dir = os.path.join(root_dir, 'temp')
    package_temp_file_prefix = f"ta-lib-{version}"
    package_temp_dir = os.path.join(temp_dir, package_temp_file_prefix)
    package_temp_file_src = os.path.join(root_dir, f"{package_temp_file_prefix}.tar.gz")
    package_temp_file_dest = os.path.join(temp_dir, f"{package_temp_file_prefix}.tar.gz")

    force_delete(package_temp_file_src, sudo_pwd)
    force_delete(package_temp_file_dest, sudo_pwd)
    force_delete(package_temp_dir, sudo_pwd)

    # Delete previous package digest files.
    digests_dir = os.path.join(dist_dir, 'digests')
    delete_other_versions(digests_dir, "*.digest", version)

    if is_build_skipping_allowed(root_dir, asset_file_name, version, sources_digest, builder_id):
        result["build_valid"] = True
        result["existed"] = True
        result["copied"] = False
        return result

    try:
        subprocess.run(['autoreconf', '-fi'], check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error running 'autoreconf -fi': {e}")
        return result

    # Run ./configure
    try:
        subprocess.run(['./configure'], check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error running './configure': {e}")
        return result

    # Run make dist
    try:
        subprocess.run(['make', 'dist'], check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error running 'make dist': {e}")
        return result

    # Move ta-lib-{version}.tar.gz into temp directory.
    if not os.path.isfile(package_temp_file_src):
        print(f"Error: {package_temp_file_src} not found.")
        return result

    os.makedirs(temp_dir, exist_ok=True)
    os.rename(package_temp_file_src, package_temp_file_dest)

    # From this point, simulate the "end-user" dev experience
    # as if the source package was freshly downloaded.
    os.chdir(temp_dir)
    os.system(f"tar -xzf {package_temp_file_dest}")
    if not test_autotool_src(package_temp_dir, sudo_pwd):
        print("Error: Source package verification failed.")
        return result

    # Move ta-lib-{version}.tar.gz into root_dir/dist (create directory as needed)
    # at same time rename it ta-lib-{version}-src.tar.gz
    # ...
    # but do this only if the archive *content* has changed
    # (ignore metadata such as file creation time).
    dist_file = os.path.join(dist_dir, asset_file_name)
    package_existed = os.path.exists(dist_file)
    package_copied = False
    if not package_existed or not compare_tar_gz_files(package_temp_file_dest, dist_file):
        os.makedirs(dist_dir, exist_ok=True)
        os.rename(package_temp_file_dest, dist_file)
        package_copied = True

    # Some clean-up (but keep the untarred directory for debugging)
    force_delete(package_temp_file_src, sudo_pwd)
    force_delete(package_temp_file_dest, sudo_pwd)

    result["build_valid"] = True
    result["existed"] = package_existed
    result["copied"] = package_copied
    # Tests were done by test_autotool_src()
    result["ta_regtest_pass"] = True
    result["dist_test_pass"] = True
    return result

def test_autotool_src(configure_dir: str, sudo_pwd: str) -> bool:
    # Returns True on success.

    # sudo_pwd is optional.
    #
    # configure_dir is the location where the end-user is expected to
    # do './configure'.
    #
    # For this test, configure_dir must be somwhere within the TA-Lib
    # git repos (will typically be under ta-lib/temp).
    #
    # In specified configure_dir do:
    # - './configure'
    # - 'make' (verify returning zero)
    # - Run './src/tools/ta_regtest/ta_regtest' (verify returning zero)
    # - Run './src/tools/gen_code/gen_code' (verify no unexpected changes)
    # - 'sudo make install' (verify returning zero)

    original_dir = os.getcwd()
    root_dir = verify_git_repo()
    generated_files_temp_copy_1 = create_temp_dir(root_dir)
    generated_files_temp_copy_2 = create_temp_dir(root_dir)
    os.chdir(configure_dir)

    try:
        git_changed = are_generated_files_git_changed(root_dir);
        copy_file_list(configure_dir,
                       generated_files_temp_copy_1,
                       get_src_generated_files())

        # Simulate typical user installation.
        subprocess.run(['./configure'], check=True)
        subprocess.run(['make'], check=True)

        # Run its src/tools/ta_regtest/ta_regtest
        subprocess.run(['src/tools/ta_regtest/ta_regtest'], check=True)

        if not os.path.isfile('src/tools/gen_code/gen_code'):
            print("Error: src/tools/gen_code/gen_code does not exist.")
            return False

        # Re-running gen_code should not cause changes to the root directory.
        # (but do nothing if there was already git changes prior to gen_code).
        # This is just a sanity check that the script is not breaking something
        # unexpected outside of the "end-user simulated" directory.
        if not git_changed and are_generated_files_git_changed(root_dir):
            print("Error: Unexpected changes from gen_code to root_dir. Do 'git diff'")
            return False

        # Now verify if gen_code did change files unexpectably within
        # the "end-user simulated" directory.
        #
        # It should not, because we are at the point of testing the src
        # package, which should have the latest generated files version
        # and re-running gen_code should have no effect.
        copy_file_list(configure_dir,
                       generated_files_temp_copy_2,
                       get_src_generated_files())
        if not compare_dir(generated_files_temp_copy_1, generated_files_temp_copy_2):
            return False

        run_command_sudo(['make', 'install'], sudo_pwd)

    except subprocess.CalledProcessError as e:
        print(f"Error during verification: {e}")
        return False

    finally:
        os.chdir(original_dir)

    return True

def update_package_digest(root_dir: str, results: dict, sources_digest: str, builder_id: str):
    # Update the digest file with the results of the packaging.
    #
    # The digest is used by the CI to skip unecessary costly steps
    # AND validate that all steps were done prior to release.
    #
    # The digest file is a JSON file:
    #    {
    #      ....
    #      "sources_digest": "some_digest_value",
    #      "package_md5": "some_md5_value",
    #      "built_success": "True".
    #      "gen_code_pass": "Disabled",
    #      "ta_regtest_pass": "True",
    #      "dist_test_pass": "False"
    #    }
    #
    # The digest files are created in dists/digests directory.
    #
    if not results.get("processed",False):
        return # NOOP

    asset_file_name = results.get("asset_file_name","")
    if asset_file_name == "":
        print("Error: asset_file_name not found in results.") # Likely a bug.
        sys.exit(1)

    pdigest = PackageDigest.read_or_create(root_dir, asset_file_name, sources_digest, builder_id)

    # Make a copy of the digest to detect changes.
    pdigest_copy = copy.deepcopy(pdigest)
    fatal_error = False
    current_md5 = "Disabled"

    if pdigest_copy.built_success == "False" or results.get("build_valid",False) == False:
        # If the previous digest or current results indicates the build was not successful, then
        # all the new tests results should be re-initialized to Unknown.
        pdigest.clear_tests()

    # A results[] key exist for a test only if the test was executed.
    # When a test is not executed, then the corresponding digest field should not be touched here.
    # (logic in other places will force "Uknown" state when, say, the build was not successful)
    if pdigest.gen_code_pass != "Disabled":
        if "gen_code_pass" in results:
            pdigest.gen_code_pass = "True" if results["gen_code_pass"] else "False"

    if pdigest.ta_regtest_pass != "Disabled":
        if "ta_regtest_pass" in results:
            pdigest.ta_regtest_pass = "True" if results["ta_regtest_pass"] else "False"

    if pdigest.dist_test_pass != "Disabled":
        if "dist_test_pass" in results:
            pdigest.dist_test_pass = "True" if results["dist_test_pass"] else "False"

    if pdigest.package_md5 != "Disabled":
        asset_file_path = path_join(root_dir, "dist", asset_file_name)
        if not os.path.exists(asset_file_path):
            # The package was processed but dist does not exists!
            print(f"Error: {asset_file_name} not found in dist.")
            fatal_error = True
        else:
            # The dist package file exists.
            current_md5 = pdigest.calculate_md5()

            if results.get("build_valid",False):
                pdigest.built_success = "True"

                if pdigest.package_md5 == "Unknown":
                    # It is a new package digest file, so just update it.
                    pdigest.package_md5 = current_md5

                if current_md5 != pdigest.package_md5 or pdigest.sources_digest != sources_digest:
                    if not results.get("copied",False):
                        # The package is different, but not successfully copied? Likely a bug.
                        print(f"Error: {asset_file_name} unexpectadly not copied to dist.")
                        fatal_error = True
                    else:
                        # The package was built and different than before...
                        print("Success: ", asset_file_name, "digest updating.")
                        pdigest.package_md5 = current_md5
                        pdigest.sources_digest = sources_digest
            else:
                # The package was not successfully rebuilt.
                print(f"Error: {asset_file_name} not successfully build.")
                fatal_error = True


    # If a github action, then reflect that the bot user did process that digest.
    # The behavior is different for local devs for which we prefer to minimize rebuilds...
    git_user_name = get_git_user_name()
    if os.getenv('GITHUB_ACTIONS') == 'true':
        pdigest.builder_id = git_user_name

    # Double check with some simple rules:
    #   - tests should never be "pass" if the build was not successful.
    #   - If a digest says the build was successful, then the md5 and sources_digest should always match.
    if not results.get("build_valid",False):
        if pdigest.dist_test_pass == "True":
            print(f"Warning: {asset_file_name} digest says dist test pass, but build was not successful.")
            pdigest.dist_test_pass = "Unknown"

        if pdigest.gen_code_pass == "True":
            print(f"Warning: {asset_file_name} digest says gen code pass, but build was not successful.")
            pdigest.gen_code_pass = "Unknown"

        if pdigest.ta_regtest_pass == "True":
            print(f"Warning: {asset_file_name} digest says ta regtest pass, but build was not successful.")
            pdigest.ta_regtest_pass = "Unknown"

    if pdigest.built_success == "True":
        if pdigest.sources_digest != sources_digest:
            print(f"Error: {asset_file_name} digest says built was successful, but sources_digest different!")
            fatal_error = True

        if pdigest.package_md5 != current_md5:
            print(f"Error: {asset_file_name} digest says built was successful, but md5 different!")
            fatal_error = True

    if fatal_error:
        pdigest.clear_tests()

    if pdigest != pdigest_copy:
        # Writing a change, and keep track of the user that made it.
        pdigest.builder_id = git_user_name
        print(f"Info: {asset_file_name} digest file updated by {git_user_name}")
        pdigest.write()

    if fatal_error:
        sys.exit(1)

def display_package_results(results: dict):
    # Display the results returned by most packaging functions.
    asset_built = results.get("processed", False)
    asset_built_success = results.get("build_valid", False)
    asset_file_name = results.get("asset_file_name", "Unknown")
    asset_copied = results.get("copied", False)
    asset_existed = results.get("existed", False)
    if asset_built:
        if not asset_built_success:
            print(f"Error: Packaging dist/{asset_file_name} failed.")

        if asset_copied:
            if asset_existed:
                print(f"Updated dist/{asset_file_name}")
            else:
                print(f"Created dist/{asset_file_name}")
        else:
            print(f"No changes for dist/{asset_file_name}")
    else:
        print(f"Skipped {asset_file_name} build (not supported on this platform)")

def package_all_linux(root_dir: str, version: str, sources_digest: str, builder_id: str, sudo_pwd: str):
    os.chdir(root_dir)

    # For consistency, the dist/ta-lib-*-src.tar.gz are created only on Ubuntu,
    # but should be usable on other Linux distributions.
    src_tar_gz_results = {
        "build_valid": False,
        "processed": False,
        "asset_file_name": f"ta-lib-{version}-src.tar.gz"
    }

    force_build = False
    if is_ubuntu():
        asset_file_name = src_tar_gz_results["asset_file_name"]
        results = package_src_tar_gz(root_dir, asset_file_name, version, sources_digest, builder_id, sudo_pwd)
        src_tar_gz_results.update(results)
        src_tar_gz_results["processed"] = True
        if not src_tar_gz_results.get("build_valid",False):
            print(f'Error: Packaging dist/{asset_file_name} failed.')
            sys.exit(1)
        if src_tar_gz_results.get("copied", False):
            # The .tar.gz file is good at detecting if the *content* is different.
            # If any changes are detected, it will force the build
            # of all other packages.
            #
            # This is not necessary, just a nice-to-have "safety net".
            force_build = True

    # When supported by host, build DEB using CMakeLists.txt (CPack)
    deb_results_arm64 = {
        "build_valid": False,
        "processed": False,
        "asset_file_name": f"ta-lib_{version}_arm64.deb"
    }
    deb_results_amd64 = {
        "build_valid": False,
        "processed": False,
        "asset_file_name": f"ta-lib_{version}_amd64.deb"
    }
    deb_results_i386 = {
        "build_valid": False,
        "processed": False,
        "asset_file_name": f"ta-lib_{version}_i386.deb"
    }
    if is_debian_based():
        if is_arm64_toolchain_installed():
            asset_file_name = deb_results_arm64["asset_file_name"]
            results = package_deb(root_dir, asset_file_name, version, sources_digest, builder_id, sudo_pwd, "toolchain-linux-arm64.cmake", force_build)
            deb_results_arm64.update(results)
            deb_results_arm64["processed"] = True
            if not deb_results_arm64.get("build_valid",False):
                print(f'Error: Packaging dist/{asset_file_name} failed.')
                sys.exit(1)
        if is_x86_64_toolchain_installed():
            asset_file_name = deb_results_amd64["asset_file_name"]
            results = package_deb(root_dir, asset_file_name, version, sources_digest, builder_id, sudo_pwd, "toolchain-linux-x86_64.cmake", force_build)
            deb_results_amd64.update(results)
            deb_results_amd64["processed"] = True
            if not deb_results_amd64.get("build_valid",False):
                print(f'Error: Packaging dist/{asset_file_name} failed.')
                sys.exit(1)
        if is_i386_toolchain_installed():
            asset_file_name = deb_results_i386["asset_file_name"]
            results = package_deb(root_dir, asset_file_name, version, sources_digest, builder_id, sudo_pwd, "toolchain-linux-i386.cmake", force_build)
            deb_results_i386.update(results)
            deb_results_i386["processed"] = True
            if not deb_results_i386.get("build_valid",False):
                print(f'Error: Packaging dist/{asset_file_name} failed.')
                sys.exit(1)

    # When supported by host, build RPM using CMakeLists.txt (CPack)
    rpm_results = {
        "build_valid": False,
        "processed": False,
        "asset_file_name": "RPM", # Default, will change.
    }
    if is_redhat_based():
        if not is_rpmbuild_installed():
            print("Error: rpmbuild not found. RPM not created")
            sys.exit(1)
        results = package_rpm(root_dir, version, sudo_pwd)
        rpm_results.update(results)
        rpm_results["processed"] = True
        if not rpm_results.get("build_valid",False):
            print(f'Error: Packaging dist/{rpm_results["asset_file_name"]} failed.')
            sys.exit(1)


    # Reflect with digest files the operations that were
    # successfully completed with the current source code.
    #
    # This is used by CI to sometimes skip costly steps AND validate
    # that all steps were done prior to release.
    update_package_digest(root_dir, src_tar_gz_results, sources_digest, builder_id)

    update_package_digest(root_dir, deb_results_arm64, sources_digest, builder_id)
    update_package_digest(root_dir, deb_results_amd64, sources_digest, builder_id)
    update_package_digest(root_dir, deb_results_i386, sources_digest, builder_id)

    # update_package_digest(root_dir, rpm_results, sources_digest, builder_id)

    # A summary of everything that was done
    print(f"\n***********")
    print(f"* Summary *")
    print(f"***********")
    display_package_results(src_tar_gz_results)
    display_package_results(deb_results_arm64)
    display_package_results(deb_results_amd64)
    display_package_results(deb_results_i386)
    display_package_results(rpm_results)

    print(f"\nPackaging completed successfully.")

def package_windows_platform(root_dir: str, version: str, sources_digest: str, builder_id: str, platform: str) -> dict:

    vcvarsall_args = []
    if platform == "x86_64":
        vcvarsall_args = ["amd64"]
    elif platform == "x86_32":
        vcvarsall_args = ["amd64_x86"]
    elif platform == "arm_64":
        vcvarsall_args = ["amd64_arm64"]
    elif platform == "arm_32":
        vcvarsall_args = ["amd64_arm"]

    zip_asset_file_name = f'ta-lib-{version}-windows-{platform}.zip'
    msi_asset_file_name = f'ta-lib-{version}-windows-{platform}.msi'

    call_vcvarsall(root_dir, vcvarsall_args)
    results = {
        "zip_results": {
            "build_valid": False,
            "processed": False,
            "asset_file_name": zip_asset_file_name,
        },
        "msi_results": {
            "build_valid": False,
            "processed": False,
            "asset_file_name": msi_asset_file_name,
        }
    }
    zip_results = package_windows_zip(root_dir, zip_asset_file_name, version, sources_digest, builder_id)
    results["zip_results"].update(zip_results)
    results["zip_results"]["processed"] = True
    if not results["zip_results"]["build_valid"]:
        print(f'Error: Packaging dist/{zip_asset_file_name} failed')
        sys.exit(1)

    # The zip file is better at detecting if the *content* is different.
    # If any changes are detected, it will force the build
    # of the .msi file in dist.
    force_build = False
    if results["zip_results"]["processed"] and results["zip_results"]["copied"]:
        force_build = True

    # For now, skip the .msi file creation if wix is not installed.

    # Process the .msi file.
    if not is_wix_installed():
        print("Warning: WiX Toolset not found. MSI packaging skipped.")
    else:
        msi_results = package_windows_msi(root_dir, msi_asset_file_name, version, sources_digest, builder_id, force_build)
        results["msi_results"].update(msi_results)
        results["msi_results"]["processed"] = True
        if not results["msi_results"]["build_valid"]:
            print(f'Error: Packaging dist/{msi_asset_file_name} failed')
            sys.exit(1)

    return results

def package_all_windows(root_dir: str, version: str, sources_digest: str, builder_id: str):
    results_x86_64 = package_windows_platform(root_dir, version, sources_digest, builder_id, "x86_64")
    results_x86_32 = package_windows_platform(root_dir, version, sources_digest, builder_id, "x86_32")

    # TODO: More testing needed for ARM platforms.
    #results_arm_64 = package_windows_platform(root_dir, version, "arm_64")
    #results_arm_32 = package_windows_platform(root_dir, version, "arm_32")

    update_package_digest(root_dir, results_x86_64["zip_results"], sources_digest, builder_id)
    update_package_digest(root_dir, results_x86_64["msi_results"], sources_digest, builder_id)
    update_package_digest(root_dir, results_x86_32["zip_results"], sources_digest, builder_id)
    update_package_digest(root_dir, results_x86_32["msi_results"], sources_digest, builder_id)

    print(f"\n***********")
    print(f"* Summary *")
    print(f"***********")
    display_package_results(results_x86_64["zip_results"])
    display_package_results(results_x86_64["msi_results"])
    display_package_results(results_x86_32["zip_results"])
    display_package_results(results_x86_32["msi_results"])
    #display_package_results(results_arm_64["zip_results"])
    #display_package_results(results_arm_64["msi_results"])
    #display_package_results(results_arm_32["zip_results"])
    #display_package_results(results_arm_32["msi_results"])

    print(f"Packaging completed successfully.")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Test release candidate assets in 'dist'")
    parser.add_argument('-p', '--pwd', type=str, default="", help="Optional password for sudo commands")
    args = parser.parse_args()

    sudo_pwd = args.pwd
    root_dir = verify_git_repo()
    is_updated, version = sync_versions(root_dir)
    is_updated, sources_digest = sync_sources_digest(root_dir)
    host_platform = sys.platform

    # The builder_id is the git user name.
    builder_id = get_git_user_name()

    if host_platform == "linux":
        package_all_linux(root_dir, version, sources_digest, builder_id, sudo_pwd)
    elif host_platform == "win32":
        arch = platform.architecture()[0]
        if arch == '64bit':
            package_all_windows(root_dir, version, sources_digest, builder_id)
        else:
            print( f"Unsupported [{arch}]. Only 64-bits windows supported for TA-Lib development.")
    else:
        print(f"Unsupported platform [{host_platform}]. Contact TA-Lib maintainers.")
        sys.exit(1)



import hashlib
import os
import re
import sys
from typing import Tuple

from .common import expand_globs
from .files import path_join

def _read_version_info(version_file_path: str, version_pattern: str) -> dict:
    """
    Read a version string from a file.

    version_pattern must be a regular expression with a first group
    that includes MAJOR, MINOR, PATCH tokens. In other words, the
    keys must be present on the line being selected by the pattern.

    The second regex group must be the value of the token.

    Returns a dictionary with the keys 'MAJOR', 'MINOR', 'PATCH'.
    """
    version_info = {'MAJOR': None, 'MINOR': None, 'PATCH': None}

    with open(version_file_path, 'r') as file:
        for line in file:
            match = version_pattern.search(line)
            if match:
                key = match.group(1)
                value = match.group(2).strip('"')
                version_info[key] = value

    if None in version_info.values():
        print(f"Error: MAJOR, MINOR, and PATCH must be defined in {version_file_path}")
        sys.exit(1)

    return version_info

def _version_info_to_string(version_info: dict) -> str:
    """
    Convert a version information dictionary to a version string.

    The dictionary must have the keys 'MAJOR', 'MINOR', 'PATCH'.
    """
    # Verify that all keys are defined.
    if None in version_info.values():
        print(f"Error: MAJOR, MINOR, and PATCH must be defined in {version_info}")
        sys.exit(1)

    return f"{version_info.get('MAJOR', '0')}.{version_info.get('MINOR', '0')}.{version_info.get('PATCH', '0')}"

def _split_version_string(version: str) -> dict:
    """
    Split a version string into its components.

    The version string must be in the format "0.0.0"

    Returns a dictionary with the keys 'MAJOR', 'MINOR', 'PATCH'
    """
    version_info = {'MAJOR': None, 'MINOR': None, 'PATCH': None}

    dot_parts = version.split('.')
    if len(dot_parts) != 3:
        print(f"Error: version must be in format '0.0.0' got instead [{version}]")
        sys.exit(1)

    version_info['MAJOR'], version_info['MINOR'], version_info['PATCH'] = dot_parts

    return version_info

def get_version_string(root_dir: str) -> str:
    """
    Parse the file VERSION to get the version string.

    The file contains a single line with the version string "major.minor.patch"

    Exit on any error.
    """
    version_file_path = path_join(root_dir, "VERSION")

    try:
        with open(version_file_path, 'r') as version_file:
            version_string = version_file.readline().strip()
        # Validate the "0.0.0" format.
        if not re.match(r'^\d+\.\d+\.\d+$', version_string):
            print(f"Error: VERSION file must contain a valid version string in the format '0.0.0'. Got instead [{version_string}]")
            sys.exit(1)
    except Exception as e:
        print(f"Error reading version file: {e}")
        sys.exit(1)

    return version_string

def set_version_string(root_dir: str, new_version:str):
    """
    Counterpart to get_version_string() that updates the VERSION file with the
    provided new_version string.
    """
    version_file_path = path_join(root_dir, "VERSION")

    current_version = get_version_string(root_dir)

    if current_version == new_version:
        return # No changes needed. The version is already up to date.

    with open(version_file_path, 'w') as version_file:
        version_file.write(new_version + '\n')

def get_version_string_source_code(root_dir: str) -> str:
    """
    Parse the file src/ta_common/ta_version.c to build a version string.

    The file contains the following C definitions:
        #define MAJOR "0"
        #define MINOR "6"
        #define PATCH "0"

    These become the string "0.6.1".
    """

    version_file_path = path_join(root_dir, "src", "ta_common", "ta_version.c")
    version_pattern = re.compile(r'#define\s+(MAJOR|MINOR|PATCH)\s+"(.*?)"')

    version_info = _read_version_info(version_file_path, version_pattern)

    return _version_info_to_string(version_info)

def set_version_string_source_code(root_dir: str, new_version:str):
    """
    Counterpart to get_version_string_source_code() that updates the src/ta_common/ta_version.c
    """
    version_file_path = path_join(root_dir, "src", "ta_common", "ta_version.c")

    current_version = get_version_string_source_code(root_dir)

    if current_version == new_version:
        return # No changes needed. The version is already up to date.

    version_info = _split_version_string(new_version)

    # Read the ta_version.c file
    with open(version_file_path, 'r') as version_file:
        lines = version_file.readlines()

    # Update the version information in the lines
    version_pattern = re.compile(r'#define\s+(MAJOR|MINOR|PATCH)\s+"(.*?)"')
    found_keys = set()
    for i, line in enumerate(lines):
        match = version_pattern.search(line)
        if match:
            key = match.group(1)
            if key in version_info:
                value = version_info[key]
                lines[i] = f'#define {key} "{value}"\n'
                found_keys.add(key)

    # Check if all required keys were found
    if not all(k in found_keys for k in ['MAJOR', 'MINOR', 'PATCH']):
        print(f"Error: MAJOR, MINOR, and PATCH must be defined in {version_file_path}")
        sys.exit(1)

    # Write the updated lines back to the ta_version.c file
    with open(version_file_path, 'w') as version_file:
        version_file.writelines(lines)

def get_version_string_cmake(root_dir: str) -> str:
    """
    Similar to get_version_string_source_code() but parse CMakeLists.txt instead of ta_version.c

    Excerpt of the file:
        SET(TA_LIB_VERSION_MAJOR 0)
        SET(TA_LIB_VERSION_MINOR 6)
        SET(TA_LIB_VERSION_PATCH 0)
    """
    version_file_path = path_join(root_dir, "CMakeLists.txt")
    version_pattern = re.compile(r'set\s*\(\s*TA_LIB_VERSION_(MAJOR|MINOR|PATCH)\s+(\d+)\s*\)', re.IGNORECASE)

    version_info = _read_version_info(version_file_path, version_pattern)

    return _version_info_to_string(version_info)


def set_version_string_cmake(root_dir: str, new_version:str):
    """
    Counterpart to get_version_string_cmake() that updates the
    file with the provided new_version string.

    The SET(TA_LIB_VERSION_XXXXXX, VALUE) pattern must already be present in the file,
    so only the VALUE portion need to be modified.

    If a given TA_LIB_VERSION_XXXXXX is not found in the file, the function will
    fail with a sys.exit(1).

    Excerpt of the file:
        SET(TA_LIB_VERSION_MAJOR 0)
        SET(TA_LIB_VERSION_MINOR 6)
        SET(TA_LIB_VERSION_PATCH 0)

    Example of new_version: "0.12.2"
    """

    version_file_path = path_join(root_dir, "CMakeLists.txt")

    current_version = get_version_string_cmake(root_dir)

    if current_version == new_version:
        return # No changes needed. The version is already up to date.

    version_info = _split_version_string(new_version)

    # Read the CMakeLists.txt file
    with open(version_file_path, 'r') as cmake_file:
        lines = cmake_file.readlines()

    # Update the version information in the lines
    version_pattern = re.compile(r'set\s*\(\s*TA_LIB_VERSION_(MAJOR|MINOR|PATCH)\s+.*\)', re.IGNORECASE)
    found_keys = set()
    for i, line in enumerate(lines):
        match = version_pattern.search(line)
        if match:
            key = match.group(1)
            if key in version_info:
                value = version_info[key]
                lines[i] = f'set(TA_LIB_VERSION_{key} {value})\n'
                found_keys.add(key)

    # Check if all required keys were found
    if not all(k in found_keys for k in ['MAJOR', 'MINOR', 'PATCH']):
        print(f"Error: MAJOR, MINOR, and PATCH must be defined in {version_file_path}")
        sys.exit(1)

    # Write the updated lines back to the CMakeLists.txt file
    with open(version_file_path, 'w') as cmake_file:
        cmake_file.writelines(lines)

def get_version_string_spec_in(root_dir: str) -> str:
    """
    Reads the version string from the ta-lib.spec.in file.
    """
    filepath = path_join(root_dir, "ta-lib.spec.in")
    version_pattern = re.compile(r'%define\s+ta_ver\s+(\d+\.\d+\.\d+)')

    with open(filepath, 'r') as file:
        for line in file:
            match = version_pattern.search(line)
            if match:
                return match.group(1)

    raise ValueError(f"ta_ver string not found in {filepath}")

def set_version_string_spec_in(root_dir: str, new_version:str):
    """
    Updates the version string in the ta-lib.spec.in file.
    """
    current_version = get_version_string_spec_in(root_dir)
    if current_version == new_version:
        print(f"No change needed. Current version is already {new_version}.")
        return

    filepath = path_join(root_dir, "ta-lib.spec.in")
    version_pattern = re.compile(r'%define\s+ta_ver\s+(\d+\.\d+\.\d+)')

    with open(filepath, 'r') as file:
        lines = file.readlines()

    updated = False
    for i, line in enumerate(lines):
        match = version_pattern.search(line)
        if match:
            current_version = match.group(1)
            if current_version != new_version:
                lines[i] = f'%define ta_ver {new_version}\n'
                updated = True
            break

    if updated:
        with open(filepath, 'w') as file:
            file.writelines(lines)
    else:
        print(f"Error: ta_ver string not found in {filepath}")
        sys.exit(1)

def get_version_string_cargo(root_dir: str) -> str:
    """
    Parse the file rust/Cargo.toml to get the version string. Example:
      version = "0.6.4"
    """
    cargo_file_path = path_join(root_dir, "rust", "Cargo.toml")

    if not os.path.exists(cargo_file_path):
        print(f"Error: Cargo.toml not found at {cargo_file_path}")
        sys.exit(1)

    version_pattern = re.compile(r'version\s*=\s*"(\d+\.\d+\.\d+)"')

    try:
        with open(cargo_file_path, 'r') as cargo_file:
            for line in cargo_file:
                match = version_pattern.search(line)
                if match:
                    return match.group(1)

        print(f"Error: Version not found in {cargo_file_path}")
        sys.exit(1)
    except Exception as e:
        print(f"Error reading Cargo.toml file: {e}")
        sys.exit(1)

def set_version_string_cargo(root_dir: str, new_version: str):
    """
    Update the version in rust/Cargo.toml.
    """
    cargo_file_path = path_join(root_dir, "rust", "Cargo.toml")

    if not os.path.exists(cargo_file_path):
        print(f"Warning: Cargo.toml not found at {cargo_file_path}")
        return  # No action if file doesn't exist

    current_version = get_version_string_cargo(root_dir)

    if current_version == new_version:
        return  # No changes needed. The version is already up to date.

    # Read the Cargo.toml file
    with open(cargo_file_path, 'r') as cargo_file:
        lines = cargo_file.readlines()

    # Update the version information in the lines
    version_pattern = re.compile(r'version\s*=\s*"(\d+\.\d+\.\d+)"')
    updated = False

    for i, line in enumerate(lines):
        match = version_pattern.search(line)
        if match:
            lines[i] = re.sub(version_pattern, f'version = "{new_version}"', line)
            updated = True
            break

    # Check if version was found and updated
    if not updated:
        print(f"Warning: Version line not found in {cargo_file_path}")
        return

    # Write the updated lines back to the Cargo.toml file
    with open(cargo_file_path, 'w') as cargo_file:
        cargo_file.writelines(lines)

def compare_version(version1: str, version2: str) -> int:
    """
    Compare two version strings.

    The format is 0.0.0

    Returns:
        -1 if version1 < version2
         0 if version1 == version2
         1 if version1 > version2
    """

    # Compare the parts as integer values
    v1_parts = list(map(int, version1.split('.')))
    v2_parts = list(map(int, version2.split('.')))
    for v1_part, v2_part in zip(v1_parts, v2_parts):
        if v1_part > v2_part:
            return 1
        elif v1_part < v2_part:
            return -1

    return 0 # Identical versions


def sync_versions(root_dir: str) -> Tuple[bool,str]:
    """
    Synchronize the version between:
          src/ta_common/ta_version.c
          rust/Cargo.toml
          CMakeLists.txt (root of repos)
          VERSION file (root of repos)

    The versions are first read from all. The highest version is selected.

    If the versions are all the same, this function will touch nothing.

    When a file has a lower version, it is updated with the highest version.

    Return true if any file was updated.
    """
    version_file = get_version_string(root_dir)
    version_c = get_version_string_source_code(root_dir)
    version_cmake = get_version_string_cmake(root_dir)
    version_spec_in = get_version_string_spec_in(root_dir)
    version_cargo = get_version_string_cargo(root_dir)

    # Identify the highest version among all sources.
    # Put the highest in the variable highest_version
    highest_version = version_file
    if compare_version(highest_version, version_cmake) < 0:
        highest_version = version_cmake
    if compare_version(highest_version, version_c) < 0:
        highest_version = version_c
    if compare_version(highest_version, version_spec_in) < 0:
        highest_version = version_spec_in
    if compare_version(highest_version, version_cargo) < 0:
        highest_version = version_cargo

    # Update files with a lower version.
    is_updated = False
    compare_result: int = compare_version(highest_version, version_file)
    if compare_result > 0:
        print(f"Updating VERSION to [{highest_version}]")
        set_version_string(root_dir, highest_version)
        is_updated = True

    compare_result: int = compare_version(highest_version, version_c)
    if compare_result > 0:
        print(f"Updating ta_version.c to [{highest_version}]")
        set_version_string_source_code(root_dir, highest_version)
        is_updated = True

    compare_result: int = compare_version(highest_version, version_cmake)
    if compare_result > 0:
        print(f"Updating CMakeLists.txt to [{highest_version}]")
        set_version_string_cmake(root_dir, highest_version)
        is_updated = True

    compare_result: int = compare_version(highest_version, version_spec_in)
    if compare_result > 0:
        print(f"Updating ta-lib.spec.in to [{highest_version}]")
        set_version_string_spec_in(root_dir, highest_version)
        is_updated = True

    compare_result: int = compare_version(highest_version, version_cargo)
    if compare_result > 0:
        print(f"Updating Cargo.toml to [{highest_version}]")
        set_version_string_cargo(root_dir, highest_version)
        is_updated = True

    return is_updated, version_c

def check_versions(root_dir: str) -> str:
    # Similar to sync_versions() but only checks if the versions are in sync, do not modify anything.
    version_file = get_version_string(root_dir)
    version_c = get_version_string_source_code(root_dir)
    version_cmake = get_version_string_cmake(root_dir)
    version_spec_in = get_version_string_spec_in(root_dir)

    if version_file != version_c:
        print(f"Error: VERSION [{version_file}] does not match ta_version.c [{version_c}]")
        return None

    if version_file != version_cmake:
        print(f"Error: VERSION [{version_file}] does not match CMakeLists.txt [{version_cmake}]")
        return None

    if version_file != version_spec_in:
        print(f"Error: VERSION [{version_file}] does not match ta-lib.spec.in [{version_spec_in}]")
        return None

    return version_file

def write_sources_digest(root_dir: str, new_digest: str) -> bool:
    """Update the ta_common.h file with the new digest."""
    ta_common_h_path = path_join(root_dir, 'include', 'ta_common.h')
    updated = False
    try:
        with open(ta_common_h_path, 'r') as f:
            lines = f.readlines()

        with open(ta_common_h_path, 'w') as f:
            for line in lines:
                if line.startswith('#define TA_LIB_SOURCES_DIGEST'):
                    if line.strip() != f'#define TA_LIB_SOURCES_DIGEST {new_digest}':
                        f.write(f'#define TA_LIB_SOURCES_DIGEST {new_digest}\n')
                        updated = True
                    else:
                        f.write(line)
                else:
                    f.write(line)
            if not updated:
                print(f"Error: Missing TA_LIB_SOURCES_DIGEST in ta_common.h")
                sys.exit(1)

        # Read it back to confirm the change
        written_digest = read_sources_digest(root_dir)
        if written_digest != new_digest:
            print(f"Error: Failed to update TA_LIB_SOURCES_DIGEST in ta_common.h")
            sys.exit(1)

    except IOError as e:
        print(f"Error updating file {ta_common_h_path}: {e}")
        raise
    return updated

def read_sources_digest(root_dir: str) -> str:
    # Read the value of the TA_LIB_SOURCES_DIGEST. Return None if not found.
    ta_common_h_path = path_join(root_dir, 'include', 'ta_common.h')
    try:
        with open(ta_common_h_path, 'r') as f:
            for line in f:
                if line.startswith('#define TA_LIB_SOURCES_DIGEST'):
                    return line.split()[2].strip()
    except IOError as e:
        print(f"Error reading file {ta_common_h_path}: {e}")
        sys.exit(1)

    return None

def calculate_sources_digest(root_dir: str, silent: bool = False) -> str:
    # This is for a calculated digest of all source file contant relevant
    # to packaging. It helps to trig CI repackaging when a change
    # is detected.
    #
    # The digest is written as a "#define TA_LIB_SOURCE_DIGEST XXXXXXXX"
    # in ta_common.h
    #
    # Return true when ta_common.h was updated with a new digest.
    file_patterns = [
        "CMakeLists.txt",
        "configure.ac",
        "ta-lib.*.in",
        "cmake/*",
        "include/ta_abstract.h",
        "include/ta_def.h",
        "include/ta_func.h",
        "include/ta_libc.h",
        "src/**/*.c",
        "src/**/*.h",
        "src/**/*.am",
        "*.am",
        "src/ta_abstract/templates/*",
        "ta_func_api.xml",
        "ta_func_list.txt",
        "java/src/**/*.java",
        "LICENSE",
        "VERSION",
    ]

    file_list = expand_globs(root_dir, file_patterns)

    # Remove potential duplicate entries
    file_list = list(set(file_list))

    # Normalize file paths by removing the root directory prefix (for portable sorting).
    normalized_file_list = [os.path.relpath(file_path, root_dir) for file_path in file_list]

    # Sort lower case to ensure consistent across platforms
    sorted_files = sorted(normalized_file_list, key=lambda x: x.lower())

    running_hash = hashlib.md5()
    n_lines = 0
    n_files = 0
    n_opens = 0
    n_chars = 0
    for file_path in sorted_files:
        try:
            n_files += 1
            full_file_path = path_join(root_dir, file_path)
            with open(full_file_path, 'r', encoding='utf-8') as f:
                n_opens += 1
                for line in f:
                    # Normalize line endings to Unix-style LF, remove leading/trailing whitespace
                    normalized_line = line.replace('\r\n', '\n').replace('\r', '\n').strip()
                    utf8_line = normalized_line.encode('utf-8')
                    n_chars += len(utf8_line)
                    running_hash.update(normalized_line.encode('utf-8'))
                    n_lines += 1
                #print(f" Hash: {running_hash.hexdigest()}, File: {file_path}, Lines: {n_lines}, Opens: {n_opens}, Chars: {n_chars}")
        except Exception as e:
            print(f"Error reading file while updating source digest [{file_path}]: {e}")
            sys.exit(1)

    if not silent:
        print(f"Digest input is n_files: {n_files}, n_lines: {n_lines}, n_opens: {n_opens}, n_chars: {n_chars}")
    return running_hash.hexdigest()

def sync_sources_digest(root_dir: str) -> Tuple[bool,str]:

    calculated_digest = calculate_sources_digest(root_dir, silent=True)
    # Update ta_common.h (touch only if different)
    current_digest = read_sources_digest(root_dir)
    if current_digest == calculated_digest:
        return False, calculated_digest

    print(f"Difference detected in source digest. Updating ta_common.h")
    print(f"Old source digest: {current_digest}")
    print(f"New source digest: {calculated_digest}")

    write_sources_digest(root_dir, calculated_digest)
    return True, calculated_digest


def check_sources_digest(root_dir: str) -> str:
    # Similar to sync_sources_digest() but only checks if the digests are in sync, do not modify anything.
    current_digest = read_sources_digest(root_dir)
    if current_digest is None:
        print(f"Error: TA_LIB_SOURCES_DIGEST not found in ta_common.h")
        return None

    calculated_digest = calculate_sources_digest(root_dir, silent=True)
    if current_digest != calculated_digest:
        print(f"Error: TA_LIB_SOURCES_DIGEST [{current_digest}] does not match calculated digest [{calculated_digest}]")
        return None

    return calculated_digest




#!/usr/bin/env python3

# This is called from Github Actions to:
#  - Detect VERSION inconsistenties among the files.
#  - Detect if some release candidate assets are missing in dist/
#  - Detect if a top entry for VERSION is missing in CHANGELOG.md
#  - Verify the top entry in CHANGELOG.md has a valid YYYY-MM-DD date within
#    one day of wallclock.
#  - Verify that all dist/digests files match the current source digests. This
#    is to ensure that various packagings and tests were all performed with the
#    current source code.
#
# If no problem are found, the script will create a temp/DRAFT_RELEASE_NOTES.md
# which is intended for the CI when creating the initial draft release.
#
# You can also call this script locally for early detection of problems that
# would prevent the release to proceed smoothly.

import argparse
import os
import re

from utilities.files import path_join
from utilities.common import get_release_assets, verify_git_repo
from utilities.versions import check_sources_digest, check_versions
from datetime import datetime

if __name__ == "__main__":

    root_dir = verify_git_repo()
    print(f"Running pre-release checks")

    parser = argparse.ArgumentParser(description='Pre-release checks')
    args = parser.parse_args()

    version = check_versions(root_dir)
    if not version:
        print("Error: Version inconsistencies found. Did you forget to run scripts/sync.py?")
        exit(1)

    sources_digest = check_sources_digest(root_dir)
    if not sources_digest:
        print("Error: Source digest inconsistencies found. Did you forget to run scripts/sync.py?")
        exit(1)

    expected_assets = get_release_assets(version)
    missing_assets = [asset for asset in expected_assets if not os.path.exists(path_join(root_dir, 'dist', asset))]
    if missing_assets:
        print("Error: Missing assets in dist/:")
        for asset in missing_assets:
            print(f"  - {asset}")

        print("Did you forget to wait for all release candidates assets be auto-generated in the dev branch?")
        exit(1)

    # Verify that the directory dist/digests already exists and contains at least one file per package
    # in dist/. The digest filename is <package_filename>.digest.
    #
    # Example: for dist/ta-lib-0.4.0.tar.gz, the source digest is dist/digests/ta-lib-0.4.0.tar.gz.digest
    dist_dir = path_join(root_dir, 'dist')
    digests_dir = path_join(dist_dir, 'digests')
    if not os.path.exists(digests_dir):
        print(f"Error: Missing {digests_dir} directory.")
        exit(1)
    # Iterate all files in dist/ and check if a digest exists for each.
    for asset in os.listdir(dist_dir):
        if os.path.isdir(path_join(dist_dir, asset)):
            continue
        digest_file = path_join(digests_dir, f"{asset}.digest")
        if not os.path.exists(digest_file):
            print(f"Error: Missing file [{asset}.digest]. Did you forget some re-build and/or tests steps?")
            exit(1)
        # Verify the digest file is for the current source code.
        with open(digest_file, 'r') as f:
            digest = f.read().strip()
            if digest != sources_digest:
                print(f"Error: Digest mismatch for [{asset}.digest]. Did you forget some re-build and/or tests steps?");
                exit(1)

    # Verify CHANGELOG.md exists and there is a top entry matching the VERSION file.
    changelog_path = path_join(root_dir, 'CHANGELOG.md')
    version_pattern = re.compile(r'##\s+\[\d+\.\d+\.\d+\].*')
    top_version_found = False
    release_notes = []

    with open(changelog_path, 'r') as f:
        for line in f:
            if version_pattern.match(line):
                if not top_version_found:
                    # Extract the 0.0.0 part of the pattern
                    changelog_version = line.split('[')[1].split(']')[0]
                    if changelog_version != version:
                        print(f"Error: Found top entry for version {changelog_version}, expected {version}.")
                        print("Did you forget to update CHANGELOG.md?")
                        exit(1)
                    top_version_found = True
                    # Extract the YYYY-MM-DD part of the pattern.
                    date = line.split(']')[1].strip()
                    if not re.match(r'\d{4}-\d{2}-\d{2}', date):
                        print(f"Error: Invalid date found in top entry of CHANGELOG.md: {date}")
                        exit(1)
                    # Verify date is within one day of wallclock.
                    # This is to ensure the release notes are up-to-date.
                    # We allow a small margin of error to account for timezone differences.
                    today = datetime.today().strftime('%Y-%m-%d')
                    if date != today:
                        print(f"Error: Invalid date found in top entry of CHANGELOG.md: {date}")
                        print(f"Expected date to be within one day from: {today}")
                        exit(1)
                    print(f"Found valid CHANGELOG.md entry: [{version}] {date}")

                else:
                    break
            # Skip from writing the header lines "## Changelog" in the release notes.
            if line.startswith('# Changelog'):
                continue

            release_notes.append(line)
    if not top_version_found:
        print(f"Error: No entry found in CHANGELOG.md for version {version}.")
        print("Did you forget to update CHANGELOG.md?")
        exit(1)

    # Write the release_notes in temp/DRAFT_RELEASE_NOTES.md
    temp_dir = path_join(root_dir, 'temp')
    os.makedirs(temp_dir, exist_ok=True)
    release_notes_path = path_join(temp_dir, 'DRAFT_RELEASE_NOTES.md')
    with open(release_notes_path, 'w') as f:
        f.write(''.join(release_notes))

    # Verify that the release notes file was indeed written.
    if not os.path.exists(release_notes_path):
        print(f"Error: Failed to write draft release notes to {release_notes_path}.")
        exit(1)

    print(f"Draft release notes written to {release_notes_path}")

    print("pre-release checks completed successfully.")



#!/usr/bin/env python3

# Test release candidate assets in 'dist'
#
# This script can be called directly by a dev and also from various
# GitHub Actions, branches and platforms (e.g. ubuntu-latest, windows-2022)
#
# The script must be executed from within a TA-Lib Git repos.
#
# Failing this test will **block** the official release.
#
# Returns a non-zero exit code if any problem is found.

import argparse
import os
import sys
import platform
import tempfile

from utilities.versions import get_version_string, get_version_string_cmake
from utilities.common import verify_git_repo,create_temp_dir
from install_tests.python import test_python_windows, test_python_linux

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Test release candidate assets in 'dist'")
    parser.add_argument('-p', '--pwd', type=str, default="", help="Optional password for sudo commands")
    args = parser.parse_args()

    sudo_pwd = args.pwd

    root_dir = verify_git_repo()

    temp_dir = create_temp_dir(root_dir)

    # Verify that CMakelists.txt and ta_version.c are in sync.
    version = get_version_string(root_dir)
    version_cmake = get_version_string_cmake(root_dir)
    if version != version_cmake:
        print(f"Version mismatch: ta_version.c [{version}] vs CMakeLists.txt [{version_cmake}]")
        print(f"Run 'scripts/package.py' to sync the version numbers in both files.")
        sys.exit(1)

    # Identify the dist package to test by this host.
    host_platform = sys.platform
    if host_platform == "linux":
        package_file_path = os.path.join(root_dir, "dist", f"ta-lib-{version}-src.tar.gz")
    elif host_platform == "win32":
        arch = platform.architecture()[0]
        if arch == '64bit':
            package_file_path = os.path.join(root_dir, "dist", f"ta-lib-{version}-windows-x86_64.zip")
        else:
            print( f"Architecture [{arch}] not yet supported. Only 64 bits supported on windows.")
    else:
        print(f"Unsupported platform [{host_platform}]")
        sys.exit(1)

    if not os.path.isfile(package_file_path):
        print(f"Package file not found: {package_file_path}. Do './scripts/package.py")
        sys.exit(1)

    # Simulate user doing a ta-lib-python installation.
    if host_platform == "linux":
        test_python_linux(package_file_path, temp_dir, version, sudo_pwd)
    elif host_platform == "win32":
        test_python_windows(package_file_path, temp_dir, version, sudo_pwd)
    else:
        print(f"Unsupported platform [{host_platform}]")
        sys.exit(1)



import asyncio
import pytafast

async def SMA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.SMA, *args, **kwargs)

async def EMA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.EMA, *args, **kwargs)

async def RSI(*args, **kwargs):
    return await asyncio.to_thread(pytafast.RSI, *args, **kwargs)

async def MACD(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MACD, *args, **kwargs)

async def BBANDS(*args, **kwargs):
    return await asyncio.to_thread(pytafast.BBANDS, *args, **kwargs)

async def ATR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ATR, *args, **kwargs)

async def ADX(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ADX, *args, **kwargs)

async def CCI(*args, **kwargs):
    return await asyncio.to_thread(pytafast.CCI, *args, **kwargs)

async def OBV(*args, **kwargs):
    return await asyncio.to_thread(pytafast.OBV, *args, **kwargs)

async def ROC(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ROC, *args, **kwargs)

async def STOCH(*args, **kwargs):
    return await asyncio.to_thread(pytafast.STOCH, *args, **kwargs)

async def MOM(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MOM, *args, **kwargs)

async def STDDEV(*args, **kwargs):
    return await asyncio.to_thread(pytafast.STDDEV, *args, **kwargs)

async def WILLR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.WILLR, *args, **kwargs)

async def NATR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.NATR, *args, **kwargs)

async def MFI(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MFI, *args, **kwargs)

async def CMO(*args, **kwargs):
    return await asyncio.to_thread(pytafast.CMO, *args, **kwargs)

async def DX(*args, **kwargs):
    return await asyncio.to_thread(pytafast.DX, *args, **kwargs)

async def MINUS_DI(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MINUS_DI, *args, **kwargs)

async def MINUS_DM(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MINUS_DM, *args, **kwargs)

async def PLUS_DI(*args, **kwargs):
    return await asyncio.to_thread(pytafast.PLUS_DI, *args, **kwargs)

async def PLUS_DM(*args, **kwargs):
    return await asyncio.to_thread(pytafast.PLUS_DM, *args, **kwargs)

async def APO(*args, **kwargs):
    return await asyncio.to_thread(pytafast.APO, *args, **kwargs)

async def AROON(*args, **kwargs):
    return await asyncio.to_thread(pytafast.AROON, *args, **kwargs)

async def AROONOSC(*args, **kwargs):
    return await asyncio.to_thread(pytafast.AROONOSC, *args, **kwargs)

async def PPO(*args, **kwargs):
    return await asyncio.to_thread(pytafast.PPO, *args, **kwargs)

async def TRIX(*args, **kwargs):
    return await asyncio.to_thread(pytafast.TRIX, *args, **kwargs)

async def ULTOSC(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ULTOSC, *args, **kwargs)

async def DEMA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.DEMA, *args, **kwargs)

async def KAMA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.KAMA, *args, **kwargs)

async def MA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MA, *args, **kwargs)

async def T3(*args, **kwargs):
    return await asyncio.to_thread(pytafast.T3, *args, **kwargs)

async def TEMA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.TEMA, *args, **kwargs)

async def TRIMA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.TRIMA, *args, **kwargs)

async def WMA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.WMA, *args, **kwargs)

async def SAR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.SAR, *args, **kwargs)

async def TRANGE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.TRANGE, *args, **kwargs)

async def AVGPRICE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.AVGPRICE, *args, **kwargs)

async def MEDPRICE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MEDPRICE, *args, **kwargs)

async def TYPPRICE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.TYPPRICE, *args, **kwargs)

async def WCLPRICE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.WCLPRICE, *args, **kwargs)

async def MIDPOINT(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MIDPOINT, *args, **kwargs)

async def MIDPRICE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MIDPRICE, *args, **kwargs)

async def ADXR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ADXR, *args, **kwargs)

async def BOP(*args, **kwargs):
    return await asyncio.to_thread(pytafast.BOP, *args, **kwargs)

async def MACDEXT(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MACDEXT, *args, **kwargs)

async def MACDFIX(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MACDFIX, *args, **kwargs)

async def STOCHF(*args, **kwargs):
    return await asyncio.to_thread(pytafast.STOCHF, *args, **kwargs)

async def STOCHRSI(*args, **kwargs):
    return await asyncio.to_thread(pytafast.STOCHRSI, *args, **kwargs)

async def ROCP(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ROCP, *args, **kwargs)

async def ROCR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ROCR, *args, **kwargs)

async def ROCR100(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ROCR100, *args, **kwargs)

async def AD(*args, **kwargs):
    return await asyncio.to_thread(pytafast.AD, *args, **kwargs)

async def ADOSC(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ADOSC, *args, **kwargs)

async def BETA(*args, **kwargs):
    return await asyncio.to_thread(pytafast.BETA, *args, **kwargs)

async def CORREL(*args, **kwargs):
    return await asyncio.to_thread(pytafast.CORREL, *args, **kwargs)

async def LINEARREG(*args, **kwargs):
    return await asyncio.to_thread(pytafast.LINEARREG, *args, **kwargs)

async def LINEARREG_ANGLE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.LINEARREG_ANGLE, *args, **kwargs)

async def LINEARREG_INTERCEPT(*args, **kwargs):
    return await asyncio.to_thread(pytafast.LINEARREG_INTERCEPT, *args, **kwargs)

async def LINEARREG_SLOPE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.LINEARREG_SLOPE, *args, **kwargs)

async def TSF(*args, **kwargs):
    return await asyncio.to_thread(pytafast.TSF, *args, **kwargs)

async def VAR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.VAR, *args, **kwargs)

async def AVGDEV(*args, **kwargs):
    return await asyncio.to_thread(pytafast.AVGDEV, *args, **kwargs)

async def ADD(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ADD, *args, **kwargs)

async def SUB(*args, **kwargs):
    return await asyncio.to_thread(pytafast.SUB, *args, **kwargs)

async def MULT(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MULT, *args, **kwargs)

async def DIV(*args, **kwargs):
    return await asyncio.to_thread(pytafast.DIV, *args, **kwargs)

async def ACOS(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ACOS, *args, **kwargs)

async def ASIN(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ASIN, *args, **kwargs)

async def ATAN(*args, **kwargs):
    return await asyncio.to_thread(pytafast.ATAN, *args, **kwargs)

async def CEIL(*args, **kwargs):
    return await asyncio.to_thread(pytafast.CEIL, *args, **kwargs)

async def COS(*args, **kwargs):
    return await asyncio.to_thread(pytafast.COS, *args, **kwargs)

async def COSH(*args, **kwargs):
    return await asyncio.to_thread(pytafast.COSH, *args, **kwargs)

async def EXP(*args, **kwargs):
    return await asyncio.to_thread(pytafast.EXP, *args, **kwargs)

async def FLOOR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.FLOOR, *args, **kwargs)

async def LN(*args, **kwargs):
    return await asyncio.to_thread(pytafast.LN, *args, **kwargs)

async def LOG10(*args, **kwargs):
    return await asyncio.to_thread(pytafast.LOG10, *args, **kwargs)

async def SIN(*args, **kwargs):
    return await asyncio.to_thread(pytafast.SIN, *args, **kwargs)

async def SINH(*args, **kwargs):
    return await asyncio.to_thread(pytafast.SINH, *args, **kwargs)

async def SQRT(*args, **kwargs):
    return await asyncio.to_thread(pytafast.SQRT, *args, **kwargs)

async def TAN(*args, **kwargs):
    return await asyncio.to_thread(pytafast.TAN, *args, **kwargs)

async def TANH(*args, **kwargs):
    return await asyncio.to_thread(pytafast.TANH, *args, **kwargs)

async def MAX(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MAX, *args, **kwargs)

async def MIN(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MIN, *args, **kwargs)

async def SUM(*args, **kwargs):
    return await asyncio.to_thread(pytafast.SUM, *args, **kwargs)

async def MINMAX(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MINMAX, *args, **kwargs)

async def MINMAXINDEX(*args, **kwargs):
    return await asyncio.to_thread(pytafast.MINMAXINDEX, *args, **kwargs)

async def HT_DCPERIOD(*args, **kwargs):
    return await asyncio.to_thread(pytafast.HT_DCPERIOD, *args, **kwargs)

async def HT_DCPHASE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.HT_DCPHASE, *args, **kwargs)

async def HT_PHASOR(*args, **kwargs):
    return await asyncio.to_thread(pytafast.HT_PHASOR, *args, **kwargs)

async def HT_SINE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.HT_SINE, *args, **kwargs)

async def HT_TRENDLINE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.HT_TRENDLINE, *args, **kwargs)

async def HT_TRENDMODE(*args, **kwargs):
    return await asyncio.to_thread(pytafast.HT_TRENDMODE, *args, **kwargs)

# Generate async wrappers for all candlestick patterns
def _make_async_cdl(name):
    async def wrapper(*args, **kwargs):
        return await asyncio.to_thread(getattr(pytafast, name), *args, **kwargs)
    wrapper.__name__ = name
    return wrapper

for _name in pytafast._CDL_STANDARD + list(pytafast._CDL_PENETRATION.keys()):
    globals()[_name] = _make_async_cdl(_name)



#!/usr/bin/env python3

# Merge remote ta-lib-temp fork (main) into the "official" ta-lib repository (dev branch)
#
# Must be executed while current directory is in the official ta-lib repository.
#
# Optionally, a "--squash <comment>" allows to reduce all changes to a single commit.
#
# How to define the URL of the fork?
# With 'git remote add ta-lib-temp'. Example:
#   git remote add ta-lib-temp https://github.com/TA-Lib/ta-lib-temp.git
#
# Why this script?
# The ta-lib-temp fork is useful while developing/testing Github actions.
# It reduces the notification noise and risk while debugging the CI.
#
# (e.g. "nobody" watch a fork, while >20 watches are on the official ta-lib).

import argparse
import subprocess
import sys
from utilities.common import verify_git_repo

def run_command(command):
    """Run a shell command and return the output."""
    result = subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    return result.stdout.strip()

def main():
    parser = argparse.ArgumentParser(description="Merge ta-lib-temp fork into the upstream ta-lib repository.")
    parser.add_argument('--squash', type=str, help="Squash all changes into a single commit with the specified comment.")
    args = parser.parse_args()


    try:
        # Verify that the current directory in within the official ta-lib repos.
        root_dir = verify_git_repo()
        remote_url = run_command(['git', 'remote', 'get-url', 'origin'])
        if not remote_url.endswith('ta-lib.git'):
            print("This script must be run from the official ta-lib repository.")
            sys.exit(1)

        # Check if ta-lib-temp is a remote
        remotes = run_command(['git', 'remote'])
        if 'ta-lib-temp' not in remotes:
            print(f"The remote ta-lib-temp is not configured.")
            print(f"Use 'git remote add ta-lib-temp <URL>' to add it.")
            print(f"Example: git remote add ta-lib-temp https://github.com/TA-Lib/ta-lib-temp.git")
            sys.exit(1)

        # Fetch the latest changes from the ta-lib-temp fork
        run_command(['git', 'fetch', 'ta-lib-temp'])

        # Checkout the main branch of the official repository
        run_command(['git', 'checkout', 'dev'])

        # Merge the cahnges from ta-lib-temp
        if args.squash:
            # Squash merge
            run_command(['git', 'merge', '--squash', 'ta-lib-temp/main'])
            run_command(['git', 'commit', '-m', args.squash])
        else:
            # Regular merge
            run_command(['git', 'merge', 'ta-lib-temp/main'])

        print("Merge completed successfully into official dev branch.\n")

    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3

# Normally, homebrew detects TA-Lib releases and creates a PR without intervention.
#
# This script is for when manual intervention is needed.
#
import argparse
from dataclasses import dataclass
import json
import os
import re
import sys
import requests

from utilities.files import calculate_file_sha256, path_join
from utilities.common import is_brew_installed, run_command, verify_git_repo_original

try:
    from github import Github, Auth
except ImportError:
    print("This script requires the PyGithub library and it is not installed. Please do 'pip install PyGithub'")
    sys.exit(1)

@dataclass
class BrewInfo:
    version: str
    sha256: str

def download_latest_github_asset_src_tar_gz(repo: any, dest_dir: str) -> str:
    # The repo parameter is the object return by PyGithub get_repo()
    latest_release = repos.get_latest_release()
    asset = None

    for a in latest_release.get_assets():
        if a.name.endswith('.tar.gz'):
            asset = a
            break

    if not asset:
        print(f"No src.tar.gz asset found")
        sys.exit(1)

    # Download the asset (unless already exists and verified same)
    asset_filepath = path_join(dest_dir, asset.name)
    if os.path.exists(asset_filepath):
        print(f"Asset already downloaded: {asset_filepath}")
        return asset.name

    asset_url = asset.browser_download_url
    os.makedirs(dest_dir, exist_ok=True)
    with requests.get(asset_url, stream=True) as r:
        r.raise_for_status()
        with open(asset_filepath, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

    print(f"Downloaded {tarball_filename} to {dest_dir}")
    return asset.name

def is_open_pr_for_formula(repo: any, formula_name: str) -> bool:
    # The repo parameter is the object return by PyGithub get_repo()
    pulls = repo.get_pulls(state='open', base='master')
    formula_prs = [pr for pr in pulls if formula_name in pr.title]

    first_letter = formula_name[0].lower()
    for pr in formula_prs:
        files = pr.get_files()
        for file in files:
            if file.filename == f"Formula/{first_letter}/{formula_name}.rb":
                return True
    return False

def verify_tarball(tarball_path: str) -> bool:
    # Sanity check that the tarball is valid.
    try:
        output = run_command(['tar', '-tf', tarball_path])
        # Verify one of the line of the output match the pattern
        # '*/src/tools/ta_regtest/ta_regtest.c'
        # Building ta_regtest is a requirement for brew.
        if not any(re.match(r".*/src/tools/ta_regtest/ta_regtest.c", line) for line in output.splitlines()):
            print(f"Error: tarball does not contain ta_regtest.c")
            return False

        return True
    except Exception as e:
        print(f"Error: {e}")
        return False

def get_ta_lib_brew_info() -> BrewInfo:
    brew_formula = "ta-lib"
    result_json = run_command(["brew", "info", "--json", brew_formula])
    brew_version = None

    # Extract the stable version and sha256. Example of JSON output:
    #[
    #  {
    #    "name": "ta-lib",
    #    "versions": {
    #      "stable": "0.4.0",
    #     },
    #    "urls":{"stable":{checksum: "sha256"}},
    #  }
    #]
    try:
        data = json.loads(result_json)
        brew_version = data[0]["versions"]["stable"]
        if brew_version is None:
            print(f"Error: Could not find the stable version for {brew_formula}")
            sys.exit(1)

        # Verify that version is the pattern x.y.z
        if not re.match(r"\d+\.\d+\.\d+", brew_version):
            print(f"Error: Invalid version found for {brew_formula}: {brew_version}")
            sys.exit(1)

        sha256 = data[0]["urls"]["stable"]["checksum"]
        if sha256 is None:
            print(f"Error: Could not find the sha256 for {brew_formula}")
            sys.exit(1)
        return BrewInfo(version=brew_version, sha256=sha256)
    except (KeyError, IndexError, json.JSONDecodeError) as e:
        print(f"Error parsing JSON: {e}")
        sys.exit(1)

    return None

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Post-release brew formula update')
    parser.add_argument('--token', help='GitHub token for authentication', required=False)
    args = parser.parse_args()

    if os.getenv('GITHUB_ACTIONS') == 'true':
        print("This script should not be called from a GitHub Action. It should be run manually/locally")
        sys.exit(1)

    print("Normally, homebrew autmatically detects TA-Lib releases and creates a PR without intervention.")
    print("Please verify first if a PR is already pending https://github.com/Homebrew/brew/pulls")
    print("This script is for manual intervention only if somehow the automatic PR was not created.\n")
    confirm = input("Are you sure you want to run this script? (yes/NO): ")
    if confirm.lower() != 'yes':
        print("Operation cancelled by the user.")
        sys.exit(0)

    if not is_brew_installed():
        print("Error: Brew needs to be installed and be on the PATH")
        sys.exit(1)

    # if an env GITHUB_TOKEN is defined... use it.
    if not args.token and 'GITHUB_TOKEN' in os.environ:
        args.token = os.getenv('GITHUB_TOKEN')
    else:
        print("Warning: No GITHUB_TOKEN env variable defined. Rate limit may fail the script.")

    if args.token:
        print("Using GitHub token for authentication")

    brew_info = get_ta_lib_brew_info()
    if not brew_info:
        print("Error: Could not get the brew info for ta-lib")
        sys.exit(1)
    print(f"Homebrew formula version: {brew_info.version}")

    # Will exit if run from a fork.
    root_dir = verify_git_repo_original()

    try:
        g = Github(args.token) if args.token else Github()
        repos = g.get_repo("ta-lib/ta-lib")
        gh_version_obj = repos.get_latest_release()
        # Remove the 'v' prefix as needed
        gh_version = gh_version_obj.title.lstrip('v')

        print(f"Github release version  : {gh_version}")
        if brew_info.version == gh_version:
            print("Homebrew formula is up-to-date")
            print("No update needed.")
            sys.exit(0)

        print("Update to Homebrew formula needed")
        old_sha256 = brew_info.sha256
        print(f"Old sha256: {old_sha256}")

        # Download the release asset in temp/
        dest_dir = path_join(root_dir, 'temp', 'post-release-brew')
        tarball_filename = download_latest_github_asset_src_tar_gz(repos,dest_dir)
        tarball_filepath = path_join(dest_dir, tarball_filename)

        # Untar just to verify that the tarball is OK.
        # Double check for consistency that the version are matching
        # (verify in case of "race condition" with github).
        #
        # Validate with regex that the filename is "ta-lib-{version}-src.tar.gz"
        # and extract the version part.
        m = re.match(r"ta-lib-(\d+\.\d+\.\d+)-src.tar.gz", tarball_filename)
        if not m:
            print(f"Error: Unexpected tarball filename: {tarball_filename}")
            sys.exit(1)
        # extract the version part
        gh_version_downloaded = m.group(1)
        if gh_version_downloaded != gh_version:
            print(f"Error: Inconsistent versions: {gh_version_downloaded} != {gh_version}")
            sys.exit(1)


        if not verify_tarball(tarball_filepath):
            print("Error: tarball verification failed")
            sys.exit(1)
        else:
            print("Tarball verification OK")

        # Calculate the sha256 of the downloaded tarball.
        new_sha256 = calculate_file_sha256(tarball_filepath)
        print(f"New sha256: {new_sha256}")

        if old_sha256 == new_sha256:
            print("Homebrew formula src.tar.gz sha256 is same")
            print("No update needed")
            sys.exit(1)

        # Check if a ta-lib PR already exists on homebrew-core
        # (to avoid creating a duplicate PR)
        hb_repo = g.get_repo("homebrew/homebrew-core")
        if is_open_pr_for_formula(hb_repo, "ta-lib"):
            print(f"PR already pending for ta-lib formula")
            print("No update needed")
            sys.exit(0)

        print("Creating PR for ta-lib formula")

        bump_cmd = ['brew', 'bump-formula-pr', '--strict']
        #bump_cmd.extend(['--dry-run'])
        gh_url = f"https://github.com/ta-lib/ta-lib/releases/download/{gh_version_obj.title}/ta-lib-{gh_version}-src.tar.gz"
        bump_cmd.extend(['--url', gh_url])
        bump_cmd.extend(['--sha256', new_sha256])
        bump_cmd.extend(['ta-lib'])

        output = run_command(bump_cmd)
        print(output)


    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)













#!/usr/bin/env python3

# Merge dev into main branch

import subprocess
import sys
import os

import sync

def run_command(command):
    """Run a shell command and return the output."""
    result = subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    return result.stdout.strip()

def main():
    try:
        # Switch to dev branch if not already on it
        original_branch = run_command(['git', 'rev-parse', '--abbrev-ref', 'HEAD'])
        if original_branch != "dev":
            print("Switching to dev branch")
            run_command(['git', 'checkout', 'dev'])

        # Make sure the dev branch does not have uncommitted changes.
        try:
            run_command(['git', 'diff-index', '--quiet', 'HEAD', '--'])
        except subprocess.CalledProcessError:
            print("Uncommitted changes in the dev branch. Please commit or stash them before merging.")
            sys.exit(1)

        # Make sure the dev branch is up-to-date with the remote.
        run_command(['git', 'fetch', 'origin', 'dev'])
        if run_command(['git', 'rev-parse', 'dev']) != run_command(['git', 'rev-parse', 'origin/dev']):
            print("dev branch not up-to-date with remote. Do 'git pull'.")
            sys.exit(1)

        # Call sync to verify that dev is up-to-date with main.
        # This is to avoid conflicts when merging dev into main.
        sync.main()

        # Switch to main branch
        print("Switching to main branch")
        run_command(['git', 'checkout', 'main'])
        run_command(['git', 'fetch', 'origin', 'main'])

        # Proceed to merge dev into main. Detect if there are conflicts, if yes
        # give instruction to resolve them.

        # Find the common ancestor of dev and main
        merge_base = run_command(['git', 'merge-base', 'dev', 'main'])

        # Check if there are any changes from dev that are not in main
        try:
            run_command(['git', 'diff', '--quiet', merge_base, 'dev'])
            print("No changes to merge from dev to main.")
        except subprocess.CalledProcessError:
            # Perform the actual merge
            try:
                run_command(['git', 'merge', '--ff-only', 'dev'])
                print("Merged dev into main.")

                # Rebase dev to keep on same last commit (that merge that was just done).
                run_command(['git', 'checkout', 'dev'])
                run_command(['git', 'rebase', 'main'])
                run_command(['git', 'push', 'origin', 'dev'])
                run_command(['git', 'push', 'origin', 'main'])
            except subprocess.CalledProcessError:
                print("Merge failed due to conflicts.")
                print("To resolve the conflicts, follow these steps:")
                print("1. Identify conflicted files using 'git status'.")
                print("2. Resolve manually by editing conflicted files.")
                print("3. Mark conflicts as resolved using 'git add <file>'.")
                print("4. Complete merge with 'git commit' and 'push'.")
                sys.exit(1)

    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")
        sys.exit(1)

    finally:
        # Restore to the branch the user was located before running this script
        current_branch = run_command(['git', 'rev-parse', '--abbrev-ref', 'HEAD'])
        if current_branch != original_branch:
            print(f"Switching back to {original_branch} branch")
            run_command(['git', 'checkout', original_branch])

if __name__ == "__main__":
    main()


import pytest
import numpy as np
import pandas as pd
import pytafast

def test_sma_numpy():
    # 10 elements
    in_real = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0])
    
    # Calculate SMA with period 3
    # Expected: 
    # [NaN, NaN, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    out = pytafast.SMA(in_real, timeperiod=3)
    
    assert isinstance(out, np.ndarray)
    assert len(out) == 10
    
    assert np.isnan(out[0])
    assert np.isnan(out[1])
    assert out[2] == pytest.approx(2.0)
    assert out[9] == pytest.approx(9.0)

def test_sma_pandas():
    # 10 elements
    in_real = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0], index=list("abcdefghij"), name="prices")
    
    out = pytafast.SMA(in_real, timeperiod=5)
    
    assert isinstance(out, pd.Series)
    assert len(out) == 10
    assert out.name == "prices"
    assert out.index.tolist() == list("abcdefghij")
    
    assert np.isnan(out.iloc[0])
    assert np.isnan(out.iloc[3])
    # SMA5 for element 4 (value 5) is sum(1,2,3,4,5)/5 = 3
    assert out.iloc[4] == pytest.approx(3.0)

def test_empty_array():
    in_real = np.array([])
    out = pytafast.SMA(in_real, timeperiod=3)
    assert len(out) == 0

def test_invalid_input():
    in_real = np.array([[1.0, 2.0], [3.0, 4.0]])
    with pytest.raises(Exception):
        pytafast.SMA(in_real, timeperiod=3)

def test_sma_against_official_talib():
    talib = pytest.importorskip("talib")
    
    # Generate random price data
    np.random.seed(42)
    in_real = np.random.random(100) * 100
    
    # Compare with official talib
    for period in [2, 5, 14, 30]:
        official_out = talib.SMA(in_real, timeperiod=period)
        our_out = pytafast.SMA(in_real, timeperiod=period)
        
        # We use np.testing.assert_allclose to handle NaNs and float precision
        np.testing.assert_allclose(our_out, official_out, equal_nan=True)

def test_ema_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100
    for period in [5, 14, 30]:
        official_out = talib.EMA(in_real, timeperiod=period)
        our_out = pytafast.EMA(in_real, timeperiod=period)
        np.testing.assert_allclose(our_out, official_out, equal_nan=True)

def test_rsi_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100
    for period in [5, 14, 30]:
        official_out = talib.RSI(in_real, timeperiod=period)
        our_out = pytafast.RSI(in_real, timeperiod=period)
        np.testing.assert_allclose(our_out, official_out, equal_nan=True)

def test_macd_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100
    
    # default params: 12, 26, 9
    o_macd, o_signal, o_hist = talib.MACD(in_real, fastperiod=12, slowperiod=26, signalperiod=9)
    p_macd, p_signal, p_hist = pytafast.MACD(in_real, fastperiod=12, slowperiod=26, signalperiod=9)
    
    np.testing.assert_allclose(p_macd, o_macd, equal_nan=True)
    np.testing.assert_allclose(p_signal, o_signal, equal_nan=True)
    np.testing.assert_allclose(p_hist, o_hist, equal_nan=True)

def test_bbands_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100
    
    for matype_val in [0, 1]:  # SMA, EMA
        o_upper, o_middle, o_lower = talib.BBANDS(in_real, timeperiod=5, nbdevup=2.0, nbdevdn=2.0, matype=matype_val)
        p_upper, p_middle, p_lower = pytafast.BBANDS(in_real, timeperiod=5, nbdevup=2.0, nbdevdn=2.0, matype=matype_val)
        
        np.testing.assert_allclose(p_upper, o_upper, equal_nan=True)
        np.testing.assert_allclose(p_middle, o_middle, equal_nan=True)
def test_atr_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.ATR(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.ATR(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_adx_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.ADX(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.ADX(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_cci_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.CCI(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.CCI(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_obv_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    in_vol = np.random.random(100) * 1000
    o_out = talib.OBV(in_real, in_vol)
    p_out = pytafast.OBV(in_real, in_vol)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_roc_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 10]:
        o_out = talib.ROC(in_real, timeperiod=period)
        p_out = pytafast.ROC(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_stoch_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    
    for matype_val in [0, 1]:  # SMA, EMA
        o_slowk, o_slowd = talib.STOCH(in_high, in_low, in_close, fastk_period=5, slowk_period=3, slowk_matype=matype_val, slowd_period=3, slowd_matype=matype_val)
        p_slowk, p_slowd = pytafast.STOCH(in_high, in_low, in_close, fastk_period=5, slowk_period=3, slowk_matype=matype_val, slowd_period=3, slowd_matype=matype_val)
        
        np.testing.assert_allclose(p_slowk, o_slowk, equal_nan=True)
        np.testing.assert_allclose(p_slowd, o_slowd, equal_nan=True)

def test_mom_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 10, 14]:
        o_out = talib.MOM(in_real, timeperiod=period)
        p_out = pytafast.MOM(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_stddev_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 10, 14]:
        for nbdev in [1.0, 2.0]:
            o_out = talib.STDDEV(in_real, timeperiod=period, nbdev=nbdev)
            p_out = pytafast.STDDEV(in_real, timeperiod=period, nbdev=nbdev)
            np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_willr_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.WILLR(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.WILLR(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_natr_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.NATR(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.NATR(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_mfi_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    in_vol = np.random.random(100) * 1000
    for period in [5, 14]:
        o_out = talib.MFI(in_high, in_low, in_close, in_vol, timeperiod=period)
        p_out = pytafast.MFI(in_high, in_low, in_close, in_vol, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_cmo_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 14]:
        o_out = talib.CMO(in_real, timeperiod=period)
        p_out = pytafast.CMO(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_dx_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.DX(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.DX(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_minus_di_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.MINUS_DI(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.MINUS_DI(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_minus_dm_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    for period in [5, 14]:
        o_out = talib.MINUS_DM(in_high, in_low, timeperiod=period)
        p_out = pytafast.MINUS_DM(in_high, in_low, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_plus_di_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.PLUS_DI(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.PLUS_DI(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_plus_dm_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    for period in [5, 14]:
        o_out = talib.PLUS_DM(in_high, in_low, timeperiod=period)
        p_out = pytafast.PLUS_DM(in_high, in_low, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_apo_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    
    for matype_val in [0, 1]:
        o_out = talib.APO(in_real, fastperiod=12, slowperiod=26, matype=matype_val)
        p_out = pytafast.APO(in_real, fastperiod=12, slowperiod=26, matype=matype_val)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_aroon_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    for period in [5, 14]:
        o_down, o_up = talib.AROON(in_high, in_low, timeperiod=period)
        p_down, p_up = pytafast.AROON(in_high, in_low, timeperiod=period)
        np.testing.assert_allclose(p_down, o_down, equal_nan=True)
        np.testing.assert_allclose(p_up, o_up, equal_nan=True)

def test_aroonosc_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    for period in [5, 14]:
        o_out = talib.AROONOSC(in_high, in_low, timeperiod=period)
        p_out = pytafast.AROONOSC(in_high, in_low, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_ppo_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    
    for matype_val in [0, 1]:
        o_out = talib.PPO(in_real, fastperiod=12, slowperiod=26, matype=matype_val)
        p_out = pytafast.PPO(in_real, fastperiod=12, slowperiod=26, matype=matype_val)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_trix_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 10]:
        o_out = talib.TRIX(in_real, timeperiod=period)
        p_out = pytafast.TRIX(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_ultosc_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    
    o_out = talib.ULTOSC(in_high, in_low, in_close, timeperiod1=7, timeperiod2=14, timeperiod3=28)
    p_out = pytafast.ULTOSC(in_high, in_low, in_close, timeperiod1=7, timeperiod2=14, timeperiod3=28)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

# --- Batch 1: Overlap Studies ---

def test_dema_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 14, 30]:
        o_out = talib.DEMA(in_real, timeperiod=period)
        p_out = pytafast.DEMA(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_kama_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 14, 30]:
        o_out = talib.KAMA(in_real, timeperiod=period)
        p_out = pytafast.KAMA(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_ma_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for matype_val in [0, 1, 2, 3, 4, 5, 6, 8]:  # SMA, EMA, WMA, DEMA, TEMA, TRIMA, KAMA, T3
        o_out = talib.MA(in_real, timeperiod=14, matype=matype_val)
        p_out = pytafast.MA(in_real, timeperiod=14, matype=matype_val)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_t3_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 10]:
        o_out = talib.T3(in_real, timeperiod=period, vfactor=0.7)
        p_out = pytafast.T3(in_real, timeperiod=period, vfactor=0.7)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_tema_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 14, 30]:
        o_out = talib.TEMA(in_real, timeperiod=period)
        p_out = pytafast.TEMA(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_trima_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 14, 30]:
        o_out = talib.TRIMA(in_real, timeperiod=period)
        p_out = pytafast.TRIMA(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_wma_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 14, 30]:
        o_out = talib.WMA(in_real, timeperiod=period)
        p_out = pytafast.WMA(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

# --- Batch 2: SAR, TRANGE, Price Transforms ---

def test_sar_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    o_out = talib.SAR(in_high, in_low, acceleration=0.02, maximum=0.2)
    p_out = pytafast.SAR(in_high, in_low, acceleration=0.02, maximum=0.2)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_trange_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    o_out = talib.TRANGE(in_high, in_low, in_close)
    p_out = pytafast.TRANGE(in_high, in_low, in_close)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_avgprice_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_open = np.random.random(100) * 100 + 10
    in_high = in_open + np.random.random(100) * 5
    in_low = in_open - np.random.random(100) * 5
    in_close = in_open + np.random.random(100) * 2 - 1
    o_out = talib.AVGPRICE(in_open, in_high, in_low, in_close)
    p_out = pytafast.AVGPRICE(in_open, in_high, in_low, in_close)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_medprice_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    o_out = talib.MEDPRICE(in_high, in_low)
    p_out = pytafast.MEDPRICE(in_high, in_low)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_typprice_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    o_out = talib.TYPPRICE(in_high, in_low, in_close)
    p_out = pytafast.TYPPRICE(in_high, in_low, in_close)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_wclprice_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    o_out = talib.WCLPRICE(in_high, in_low, in_close)
    p_out = pytafast.WCLPRICE(in_high, in_low, in_close)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_midpoint_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 14]:
        o_out = talib.MIDPOINT(in_real, timeperiod=period)
        p_out = pytafast.MIDPOINT(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_midprice_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    for period in [5, 14]:
        o_out = talib.MIDPRICE(in_high, in_low, timeperiod=period)
        p_out = pytafast.MIDPRICE(in_high, in_low, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

# --- Batch 3: Momentum indicators ---

def test_adxr_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(200) * 100 + 10
    in_low = in_high - np.random.random(200) * 5
    in_close = in_low + (in_high - in_low) / 2
    for period in [5, 14]:
        o_out = talib.ADXR(in_high, in_low, in_close, timeperiod=period)
        p_out = pytafast.ADXR(in_high, in_low, in_close, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_bop_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_open = np.random.random(100) * 100 + 10
    in_high = in_open + np.random.random(100) * 5
    in_low = in_open - np.random.random(100) * 5
    in_close = in_open + np.random.random(100) * 2 - 1
    o_out = talib.BOP(in_open, in_high, in_low, in_close)
    p_out = pytafast.BOP(in_open, in_high, in_low, in_close)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_macdext_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    o_macd, o_signal, o_hist = talib.MACDEXT(in_real, fastperiod=12, fastmatype=0,
                                              slowperiod=26, slowmatype=0,
                                              signalperiod=9, signalmatype=0)
    p_macd, p_signal, p_hist = pytafast.MACDEXT(in_real, fastperiod=12, fastmatype=0,
                                                 slowperiod=26, slowmatype=0,
                                                 signalperiod=9, signalmatype=0)
    np.testing.assert_allclose(p_macd, o_macd, equal_nan=True)
    np.testing.assert_allclose(p_signal, o_signal, equal_nan=True)
    np.testing.assert_allclose(p_hist, o_hist, equal_nan=True)

def test_macdfix_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    o_macd, o_signal, o_hist = talib.MACDFIX(in_real, signalperiod=9)
    p_macd, p_signal, p_hist = pytafast.MACDFIX(in_real, signalperiod=9)
    np.testing.assert_allclose(p_macd, o_macd, equal_nan=True)
    np.testing.assert_allclose(p_signal, o_signal, equal_nan=True)
    np.testing.assert_allclose(p_hist, o_hist, equal_nan=True)

def test_stochf_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    o_fastk, o_fastd = talib.STOCHF(in_high, in_low, in_close,
                                      fastk_period=5, fastd_period=3, fastd_matype=0)
    p_fastk, p_fastd = pytafast.STOCHF(in_high, in_low, in_close,
                                         fastk_period=5, fastd_period=3, fastd_matype=0)
    np.testing.assert_allclose(p_fastk, o_fastk, equal_nan=True)
    np.testing.assert_allclose(p_fastd, o_fastd, equal_nan=True)

def test_stochrsi_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(200) * 100 + 10
    o_fastk, o_fastd = talib.STOCHRSI(in_real, timeperiod=14,
                                        fastk_period=5, fastd_period=3, fastd_matype=0)
    p_fastk, p_fastd = pytafast.STOCHRSI(in_real, timeperiod=14,
                                           fastk_period=5, fastd_period=3, fastd_matype=0)
    np.testing.assert_allclose(p_fastk, o_fastk, equal_nan=True)
    np.testing.assert_allclose(p_fastd, o_fastd, equal_nan=True)

def test_rocp_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 10]:
        o_out = talib.ROCP(in_real, timeperiod=period)
        p_out = pytafast.ROCP(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_rocr_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 10]:
        o_out = talib.ROCR(in_real, timeperiod=period)
        p_out = pytafast.ROCR(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_rocr100_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [5, 10]:
        o_out = talib.ROCR100(in_real, timeperiod=period)
        p_out = pytafast.ROCR100(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

# --- Batch 4: Volume + Statistics indicators ---

def test_ad_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    in_volume = np.random.random(100) * 1000000
    o_out = talib.AD(in_high, in_low, in_close, in_volume)
    p_out = pytafast.AD(in_high, in_low, in_close, in_volume)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_adosc_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_high = np.random.random(100) * 100 + 10
    in_low = in_high - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) / 2
    in_volume = np.random.random(100) * 1000000
    o_out = talib.ADOSC(in_high, in_low, in_close, in_volume, fastperiod=3, slowperiod=10)
    p_out = pytafast.ADOSC(in_high, in_low, in_close, in_volume, fastperiod=3, slowperiod=10)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_beta_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in0 = np.random.random(100) * 100 + 10
    in1 = np.random.random(100) * 100 + 10
    for period in [5, 10]:
        o_out = talib.BETA(in0, in1, timeperiod=period)
        p_out = pytafast.BETA(in0, in1, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_correl_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in0 = np.random.random(100) * 100 + 10
    in1 = np.random.random(100) * 100 + 10
    for period in [10, 30]:
        o_out = talib.CORREL(in0, in1, timeperiod=period)
        p_out = pytafast.CORREL(in0, in1, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_linearreg_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [7, 14]:
        o_out = talib.LINEARREG(in_real, timeperiod=period)
        p_out = pytafast.LINEARREG(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_linearreg_angle_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    o_out = talib.LINEARREG_ANGLE(in_real, timeperiod=14)
    p_out = pytafast.LINEARREG_ANGLE(in_real, timeperiod=14)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_linearreg_intercept_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    o_out = talib.LINEARREG_INTERCEPT(in_real, timeperiod=14)
    p_out = pytafast.LINEARREG_INTERCEPT(in_real, timeperiod=14)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_linearreg_slope_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    o_out = talib.LINEARREG_SLOPE(in_real, timeperiod=14)
    p_out = pytafast.LINEARREG_SLOPE(in_real, timeperiod=14)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_tsf_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [7, 14]:
        o_out = talib.TSF(in_real, timeperiod=period)
        p_out = pytafast.TSF(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_var_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    o_out = talib.VAR(in_real, timeperiod=5, nbdev=1.0)
    p_out = pytafast.VAR(in_real, timeperiod=5, nbdev=1.0)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_avgdev_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [7, 14]:
        o_out = talib.AVGDEV(in_real, timeperiod=period)
        p_out = pytafast.AVGDEV(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

# --- Batch 5: Math Operators + Math Transforms ---

def test_add_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in0 = np.random.random(100) * 100
    in1 = np.random.random(100) * 100
    o_out = talib.ADD(in0, in1)
    p_out = pytafast.ADD(in0, in1)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_sub_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in0 = np.random.random(100) * 100
    in1 = np.random.random(100) * 100
    o_out = talib.SUB(in0, in1)
    p_out = pytafast.SUB(in0, in1)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_mult_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in0 = np.random.random(100) * 100
    in1 = np.random.random(100) * 100
    o_out = talib.MULT(in0, in1)
    p_out = pytafast.MULT(in0, in1)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_div_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in0 = np.random.random(100) * 100
    in1 = np.random.random(100) * 100 + 1
    o_out = talib.DIV(in0, in1)
    p_out = pytafast.DIV(in0, in1)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

_MATH_TRANSFORMS = [
    "ACOS", "ASIN", "ATAN", "CEIL", "COS", "COSH", "EXP", "FLOOR",
    "LN", "LOG10", "SIN", "SINH", "SQRT", "TAN", "TANH",
]

@pytest.mark.parametrize("func_name", _MATH_TRANSFORMS)
def test_math_transform_against_official_talib(func_name):
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 0.98 + 0.01
    o_out = getattr(talib, func_name)(in_real)
    p_out = getattr(pytafast, func_name)(in_real)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

# --- Batch 6: Statistics (MIN/MAX/SUM/MINMAX/MINMAXINDEX) + Cycle (HT_*) ---

def test_max_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [10, 30]:
        o_out = talib.MAX(in_real, timeperiod=period)
        p_out = pytafast.MAX(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_min_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [10, 30]:
        o_out = talib.MIN(in_real, timeperiod=period)
        p_out = pytafast.MIN(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_sum_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    for period in [10, 30]:
        o_out = talib.SUM(in_real, timeperiod=period)
        p_out = pytafast.SUM(in_real, timeperiod=period)
        np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_minmax_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    o_min, o_max = talib.MINMAX(in_real, timeperiod=30)
    p_min, p_max = pytafast.MINMAX(in_real, timeperiod=30)
    np.testing.assert_allclose(p_min, o_min, equal_nan=True)
    np.testing.assert_allclose(p_max, o_max, equal_nan=True)

def test_minmaxindex_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.random.random(100) * 100 + 10
    o_minidx, o_maxidx = talib.MINMAXINDEX(in_real, timeperiod=30)
    p_minidx, p_maxidx = pytafast.MINMAXINDEX(in_real, timeperiod=30)
    # Compare only positions after lookback (where TA-Lib outputs valid data)
    lookback = 29  # timeperiod - 1
    np.testing.assert_array_equal(np.array(p_minidx)[lookback:], np.array(o_minidx)[lookback:].astype(int))
    np.testing.assert_array_equal(np.array(p_maxidx)[lookback:], np.array(o_maxidx)[lookback:].astype(int))

def test_ht_dcperiod_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.cumsum(np.random.random(200) - 0.5) + 100
    o_out = talib.HT_DCPERIOD(in_real)
    p_out = pytafast.HT_DCPERIOD(in_real)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_ht_dcphase_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.cumsum(np.random.random(200) - 0.5) + 100
    o_out = talib.HT_DCPHASE(in_real)
    p_out = pytafast.HT_DCPHASE(in_real)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_ht_phasor_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.cumsum(np.random.random(200) - 0.5) + 100
    o_ip, o_q = talib.HT_PHASOR(in_real)
    p_ip, p_q = pytafast.HT_PHASOR(in_real)
    np.testing.assert_allclose(p_ip, o_ip, equal_nan=True)
    np.testing.assert_allclose(p_q, o_q, equal_nan=True)

def test_ht_sine_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.cumsum(np.random.random(200) - 0.5) + 100
    o_sine, o_lead = talib.HT_SINE(in_real)
    p_sine, p_lead = pytafast.HT_SINE(in_real)
    np.testing.assert_allclose(p_sine, o_sine, equal_nan=True)
    np.testing.assert_allclose(p_lead, o_lead, equal_nan=True)

def test_ht_trendline_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.cumsum(np.random.random(200) - 0.5) + 100
    o_out = talib.HT_TRENDLINE(in_real)
    p_out = pytafast.HT_TRENDLINE(in_real)
    np.testing.assert_allclose(p_out, o_out, equal_nan=True)

def test_ht_trendmode_against_official_talib():
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_real = np.cumsum(np.random.random(200) - 0.5) + 100
    o_out = talib.HT_TRENDMODE(in_real)
    p_out = pytafast.HT_TRENDMODE(in_real)
    # Compare valid positions (after lookback)
    valid = ~np.isnan(o_out.astype(float))
    np.testing.assert_array_equal(np.array(p_out)[valid], np.array(o_out)[valid].astype(int))

# --- Batch 7: Candlestick Patterns ---

_CDL_STANDARD_LIST = [
    "CDL2CROWS", "CDL3BLACKCROWS", "CDL3INSIDE", "CDL3LINESTRIKE",
    "CDL3OUTSIDE", "CDL3STARSINSOUTH", "CDL3WHITESOLDIERS",
    "CDLADVANCEBLOCK", "CDLBELTHOLD", "CDLBREAKAWAY",
    "CDLCLOSINGMARUBOZU", "CDLCONCEALBABYSWALL", "CDLCOUNTERATTACK",
    "CDLDOJI", "CDLDOJISTAR", "CDLDRAGONFLYDOJI", "CDLENGULFING",
    "CDLGAPSIDESIDEWHITE", "CDLGRAVESTONEDOJI", "CDLHAMMER",
    "CDLHANGINGMAN", "CDLHARAMI", "CDLHARAMICROSS", "CDLHIGHWAVE",
    "CDLHIKKAKE", "CDLHIKKAKEMOD", "CDLHOMINGPIGEON",
    "CDLIDENTICAL3CROWS", "CDLINNECK", "CDLINVERTEDHAMMER",
    "CDLKICKING", "CDLKICKINGBYLENGTH", "CDLLADDERBOTTOM",
    "CDLLONGLEGGEDDOJI", "CDLLONGLINE", "CDLMARUBOZU",
    "CDLMATCHINGLOW", "CDLONNECK", "CDLPIERCING", "CDLRICKSHAWMAN",
    "CDLRISEFALL3METHODS", "CDLSEPARATINGLINES", "CDLSHOOTINGSTAR",
    "CDLSHORTLINE", "CDLSPINNINGTOP", "CDLSTALLEDPATTERN",
    "CDLSTICKSANDWICH", "CDLTAKURI", "CDLTASUKIGAP", "CDLTHRUSTING",
    "CDLTRISTAR", "CDLUNIQUE3RIVER", "CDLUPSIDEGAP2CROWS",
    "CDLXSIDEGAP3METHODS",
]

_CDL_PEN_LIST = [
    "CDLABANDONEDBABY", "CDLDARKCLOUDCOVER", "CDLEVENINGDOJISTAR",
    "CDLEVENINGSTAR", "CDLMATHOLD", "CDLMORNINGDOJISTAR", "CDLMORNINGSTAR",
]

@pytest.mark.parametrize("func_name", _CDL_STANDARD_LIST)
def test_cdl_standard_against_official_talib(func_name):
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_open = np.random.random(100) * 50 + 50
    in_high = in_open + np.random.random(100) * 5
    in_low = in_open - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) * np.random.random(100)
    o_out = getattr(talib, func_name)(in_open, in_high, in_low, in_close)
    p_out = getattr(pytafast, func_name)(in_open, in_high, in_low, in_close)
    valid = ~np.isnan(o_out.astype(float))
    np.testing.assert_array_equal(np.array(p_out)[valid], np.array(o_out)[valid].astype(int))

@pytest.mark.parametrize("func_name", _CDL_PEN_LIST)
def test_cdl_penetration_against_official_talib(func_name):
    talib = pytest.importorskip("talib")
    np.random.seed(42)
    in_open = np.random.random(100) * 50 + 50
    in_high = in_open + np.random.random(100) * 5
    in_low = in_open - np.random.random(100) * 5
    in_close = in_low + (in_high - in_low) * np.random.random(100)
    o_out = getattr(talib, func_name)(in_open, in_high, in_low, in_close)
    p_out = getattr(pytafast, func_name)(in_open, in_high, in_low, in_close)
    valid = ~np.isnan(o_out.astype(float))
    np.testing.assert_array_equal(np.array(p_out)[valid], np.array(o_out)[valid].astype(int))
